# coding=utf-8

from __future__ import division
import time
import math
import socket
import logging
import threading

from pymycobot.log import setup_logging
from pymycobot.generate import CommandGenerator
from pymycobot.common import ProtocolCode, write, read
from pymycobot.error import calibration_parameters
from pymycobot.sms import sms_sts

class MyPalletizerSocket(CommandGenerator, sms_sts):
    """MyCobot Python API Serial communication class.
    Note: Please use this class under the same network

    Supported methods:

        # Overall status
            set_free_mode()
            is_free_mode()
            Other at parent class: `CommandGenerator`.

        # MDI mode and operation
            get_radians()
            send_radians()
            sync_send_angles() *
            sync_send_coords() *
            Other look at parent class: `CommandGenerator`.

        # JOG mode and operation
            Look at parent class: `CommandGenerator`.

        # Running status and Settings
            Look at parent class: `CommandGenerator`.

        # Servo control
            Look at parent class: `CommandGenerator`.

        # Atom IO
            Look at parent class: `CommandGenerator`.

        # Basic
            Look at parent class: `CommandGenerator`.

        # Other
            wait() *
    """
    _write = write
    _read = read

    def __init__(self, ip, netport=9000, debug=False):
        """
        Args:
            ip: Server ip
            netport: Server port
        """
        super(MyPalletizerSocket, self).__init__(debug)
        self.calibration_parameters = calibration_parameters
        self.SERVER_IP = ip
        self.SERVER_PORT = netport
        self.sock = self.connect_socket()
        self.lock = threading.Lock()

    def connect_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.SERVER_IP, self.SERVER_PORT))
        return sock

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply, _async = super(
            MyPalletizerSocket, self)._mesg(genre, *args, **kwargs)
        # [254,...,255]
        with self.lock:
            data = self._write(self._flatten(real_command), "socket")
            if has_reply:
                data = self._read(genre, method='socket')
                if genre == ProtocolCode.SET_SSID_PWD:
                    return None
                res = self._process_received(data, genre)
                if res == []:
                    return None
                if genre in [
                    ProtocolCode.ROBOT_VERSION,
                    ProtocolCode.IS_POWER_ON,
                    ProtocolCode.IS_CONTROLLER_CONNECTED,
                    ProtocolCode.IS_PAUSED,
                    ProtocolCode.IS_IN_POSITION,
                    ProtocolCode.IS_MOVING,
                    ProtocolCode.IS_SERVO_ENABLE,
                    ProtocolCode.IS_ALL_SERVO_ENABLE,
                    ProtocolCode.GET_SERVO_DATA,
                    ProtocolCode.GET_DIGITAL_INPUT,
                    ProtocolCode.GET_GRIPPER_VALUE,
                    ProtocolCode.IS_GRIPPER_MOVING,
                    ProtocolCode.GET_SPEED,
                    ProtocolCode.GET_ENCODER,
                    ProtocolCode.GET_BASIC_INPUT,
                    ProtocolCode.GET_TOF_DISTANCE,
                    ProtocolCode.GET_GPIO_IN,
                    ProtocolCode.GET_COMMUNICATE_MODE,
                    ProtocolCode.SET_COMMUNICATE_MODE,
                    ProtocolCode.SetHTSGripperTorque,
                    ProtocolCode.GetHTSGripperTorque,
                    ProtocolCode.GetGripperProtectCurrent,
                    ProtocolCode.InitGripper,
                    ProtocolCode.SET_FOUR_PIECES_ZERO
                ]:
                    return self._process_single(res)
                elif genre in [ProtocolCode.GET_ANGLES]:
                    return [self._int2angle(angle) for angle in res]
                elif genre in [ProtocolCode.GET_COORDS]:
                    if res:
                        r = []
                        for idx in range(3):
                            r.append(self._int2coord(res[idx]))
                        if len(res) > 3:
                            r.append(self._int2angle(res[3]))
                        return r
                    else:
                        return res
                elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
                    return [self._int2coord(angle) for angle in res]
                elif genre in [ProtocolCode.GET_JOINT_MAX_ANGLE, ProtocolCode.GET_JOINT_MIN_ANGLE]:
                    return self._int2coord(res[0])
                elif genre in [ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION,
                               ProtocolCode.GET_ATOM_VERSION]:
                    return self._int2coord(self._process_single(res))
                elif genre == ProtocolCode.GET_ANGLES_COORDS:
                    r = []
                    for index in range(len(res)):
                        if index < 4:
                            r.append(self._int2angle(res[index]))
                        elif index < 7:
                            r.append(self._int2coord(res[index]))
                        else:
                            r.append(self._int2angle(res[index]))
                    return r
                else:
                    return res
            return None

    # Overall Status
    def set_free_mode(self, flag):
        """set to free mode

        Args:
            flag: 0/1
        """
        self.calibration_parameters(class_name=self.__class__.__name__, flag=flag)
        return self._mesg(ProtocolCode.SET_FREE_MODE, flag)

    def is_free_mode(self):
        """Check if it is free mode

        Return:
            0/1
        """
        return self._mesg(ProtocolCode.IS_FREE_MODE, has_reply=True)

    # MDI mode and operation
    def get_radians(self):
        """Get all angle return a list

        Return:
            data_list (list[radian...]):
        """
        angles = self._mesg(ProtocolCode.GET_ANGLES, has_reply=True)
        return [round(angle * (math.pi / 180), 3) for angle in angles]

    def send_radians(self, radians, speed):
        """Send all angles

        Args:
            radians (list): example [0, 0, 0, 0, 0, 0]
            speed (int): 0 ~ 100
        """
        # calibration_parameters(len6=radians, speed=speed)
        degrees = [self._angle2int(radian * (180 / math.pi))
                   for radian in radians]
        return self._mesg(ProtocolCode.SEND_ANGLES, degrees, speed)

    def send_angle(self, id, degree, speed):
        """Send one angle of joint to robot arm.

        Args:
            id : Joint id(genre.Angle) int 1-4.
            angle : angle value(float).
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=id, angle=degree, speed=speed)
        return self._mesg(ProtocolCode.SEND_ANGLE, id, [self._angle2int(degree)], speed, has_reply=True)

    def send_angles(self, angles, speed):
        """Send the angles of all joints to robot arm.

        Args:
            angles: a list of angle values(List[float]). len 4.
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, angles=angles, speed=speed)
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.SEND_ANGLES, angles, speed, has_reply=True)

    def get_coords(self):
        """Get the coords from robot arm, coordinate system based on base.

        Return:
            list : A float list of coord . [x, y, z, θ].
        """
        return self._mesg(ProtocolCode.GET_COORDS, has_reply=True)

    def send_coord(self, id, coord, speed):
        """Send one coord to robot arm.

        Args:
            id(int) : coord id(genre.Coord)\n int 1-4.
            coord(float) : coord value, mm
            speed(int) : 1 ~ 100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=id, coord=coord, speed=speed)
        value = self._coord2int(coord) if id <= 3 else self._angle2int(coord)
        return self._mesg(ProtocolCode.SEND_COORD, id, [value], speed, has_reply=True)

    def send_coords(self, coords, speed):
        """Send all coords to robot arm.

        Args:
            coords: a list of coords value(List[float]). [x, y, z, θ]
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, coords=coords, speed=speed)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SEND_COORDS, coord_list, speed, has_reply=True)

    def is_in_position(self, data, id=0):
        """Judge whether in the position.

        Args:
            data: A data list, angles or coords. len 4
            id: 1 - coords, 0 - angles

        Return:
            1 - True\n
            0 - False\n
            -1 - Error
        """
        if id == 1:
            self.calibration_parameters(class_name=self.__class__.__name__, coords=data)
            data_list = []
            for idx in range(3):
                data_list.append(self._coord2int(data[idx]))
            for idx in range(3, 4):
                data_list.append(self._angle2int(data[idx]))
        elif id == 0:
            self.calibration_parameters(class_name=self.__class__.__name__, angles=data)
            data_list = [self._angle2int(i) for i in data]
        else:
            raise Exception("id is not right, please input 0 or 1")

        return self._mesg(ProtocolCode.IS_IN_POSITION, data_list, id, has_reply=True)

    def sync_send_angles(self, degrees, speed, timeout=15):
        t = time.time()
        self.send_angles(degrees, speed)
        while time.time() - t < timeout:
            f = self.is_moving()
            if not f:
                break
            time.sleep(0.1)
        return 1

    def sync_send_coords(self, coords, speed, timeout=15):
        t = time.time()
        self.send_coords(coords, speed)
        while time.time() - t < timeout:
            if not self.is_moving():
                break
            time.sleep(0.1)
        return 1

    # JOG mode and operation
    def jog_angle(self, joint_id, direction, speed):
        """Jog control angle.
        Args:
            joint_id: int 1-4.
            direction: 0 - decrease, 1 - increase
            speed: int (0 - 100)
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=joint_id, direction=direction)
        return self._mesg(ProtocolCode.JOG_ANGLE, joint_id, direction, speed)

    def jog_coord(self, coord_id, direction, speed):
        """Jog control coord.

        Args:
            coord_id: int 1-4.
            direction: 0 - decrease, 1 - increase
            speed: int (1 - 100)
        """
        self.calibration_parameters(class_name=self.__class__.__name__, coord_id=coord_id, direction=direction)
        return self._mesg(ProtocolCode.JOG_COORD, coord_id, direction, speed)

    def jog_absolute(self, joint_id, angle, speed):
        """Jog absolute angle

        Args:
            joint_id: int 1-4.
            angle: -180 ~ 180
            speed: int (1 - 100)
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=joint_id, angle=angle, speed=speed)
        return self._mesg(ProtocolCode.JOG_ABSOLUTE, joint_id, [self._angle2int(angle)], speed)

    def jog_increment(self, joint_id, increment, speed):
        """step mode

        Args:
            joint_id: int 1-4.
            increment:
            speed: int (0 - 100)
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=joint_id, speed=speed)
        return self._mesg(ProtocolCode.JOG_INCREMENT, joint_id, [self._angle2int(increment)], speed)

    def jog_stop(self):
        """Stop jog moving"""
        return self._mesg(ProtocolCode.JOG_STOP)

    def set_encoder(self, joint_id, encoder, speed):
        """Set a single joint rotation to the specified potential value.

        Args:
            joint_id: int 1 - 4
            encoder: The value of the set encoder.
            speed : 1 - 100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, encode_id=joint_id, encoder=encoder,
                                    speed=speed)
        return self._mesg(ProtocolCode.SET_ENCODER, joint_id, [encoder], speed)

    def get_encoder(self, joint_id):
        """Obtain the specified joint potential value.

        Args:
            joint_id: (int) 1 - 4
        """
        self.calibration_parameters(class_name=self.__class__.__name__, encode_id=joint_id)
        return self._mesg(ProtocolCode.GET_ENCODER, joint_id, has_reply=True)

    def set_encoders(self, encoders, sp):
        """Set the six joints of the manipulator to execute synchronously to the specified position.

        Args:
            encoders: A encoder list. len 4
            sp: speed 1 ~ 100
        """
        return self._mesg(ProtocolCode.SET_ENCODERS, encoders, sp)

    # Running Status and Settings
    def get_speed(self):
        """
        Get speed
        """
        return self._mesg(ProtocolCode.GET_SPEED, has_reply=True)

    def set_speed(self, speed):
        """
        Set speed
        :param speed: 0-100
        """
        self.calibration_parameters(class_name=self.__class__.__name__, speed=speed)
        return self._mesg(ProtocolCode.SET_SPEED, speed)

    def get_feed_override(self):
        return self._process_single(
            self._mesg(ProtocolCode.GET_FEED_OVERRIDE, has_reply=True)
        )

    def get_acceleration(self):
        """get acceleration"""
        return self._process_single(
            self._mesg(ProtocolCode.GET_ACCELERATION, has_reply=True)
        )

    def set_acceleration(self, acc):
        """Set speed for all moves

        Args:
            acc: int
        """
        return self._mesg(ProtocolCode.SET_ACCELERATION, acc)

    def get_joint_min_angle(self, joint_id):
        """Gets the minimum movement angle of the specified joint

        Args:
            joint_id: 0 - 3
        Return:
            angle value(float)
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MIN_ANGLE, joint_id, has_reply=True)

    def get_joint_max_angle(self, joint_id):
        """Gets the maximum movement angle of the specified joint

        Args:
            joint_id: 0 - 3
        Return:
            angle value(float)
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MAX_ANGLE, joint_id, has_reply=True)

    def set_joint_min(self, id, angle):
        """Set the joint minimum angle

        Args:
            id: int 0 - 3
            angle: 0 ~ 180
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=id, angle=angle)
        return self._mesg(ProtocolCode.SET_JOINT_MIN, id, angle)

    def set_joint_max(self, id, angle):
        """Set the joint maximum angle

        Args:
            id: int 0 - 3
            angle: 0 ~ 180
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=id, angle=angle)
        return self._mesg(ProtocolCode.SET_JOINT_MAX, id, angle)

    # Servo control
    def is_servo_enable(self, servo_id):
        """To detect the connection state of a single joint

        Args:
            servo_id: Joint id 1 - 4
        Return:
            0 - disable
            1 - enable
            -1 - error
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id)
        return self._mesg(ProtocolCode.IS_SERVO_ENABLE, servo_id, has_reply=True)

    def set_servo_data(self, servo_id, data_id, value, mode=None):
        """Set the data parameters of the specified address of the steering gear

        Args:
            servo_id: Serial number of articulated steering gear. Joint id 1 - 4
            data_id: Data address.
            value: 0 - 4096
            mode: 0 - indicates that value is one byte(default), 1 - 1 represents a value of two bytes.
        """
        if mode is None:
            self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id, address=data_id,
                                        value=value)
            return self._mesg(ProtocolCode.SET_SERVO_DATA, servo_id, data_id, value)
        else:
            self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id, address=data_id,
                                        value=value,
                                        mode=mode)
            return self._mesg(ProtocolCode.SET_SERVO_DATA, servo_id, data_id, [value], mode)

    def get_servo_data(self, servo_id, data_id, mode=None):
        """Read the data parameter of the specified address of the steering gear.

        Args:
            servo_id: Serial number of articulated steering gear. Joint id 1 - 4
            data_id: Data address.
            mode: 0 - indicates that value is one byte(default), 1 - 1 represents a value of two bytes.

        Return:
            values 0 - 4096
        """
        if mode is not None:
            self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id, address=data_id, mode=mode)
            return self._mesg(
                ProtocolCode.GET_SERVO_DATA, servo_id, data_id, mode, has_reply=True
            )
        self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id, address=data_id)
        return self._mesg(
            ProtocolCode.GET_SERVO_DATA, servo_id, data_id, has_reply=True
        )

    def set_servo_calibration(self, servo_id):
        """The current position of the calibration joint actuator is the angle zero point,
            and the corresponding potential value is 2048.

        Args:
            servo_id: Serial number of articulated steering gear. Joint id 1 - 4
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id)
        return self._mesg(ProtocolCode.SET_SERVO_CALIBRATION, servo_id)

    def release_servo(self, servo_id, mode=None):
        """Power off designated servo

        Args:
            servo_id: int Joint id 1 - 4
            mode: Default damping, set to 1, cancel damping
        """
        if mode is None:
            self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id)
            return self._mesg(ProtocolCode.RELEASE_SERVO, servo_id)

        else:
            self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id)
            return self._mesg(ProtocolCode.RELEASE_SERVO, servo_id, mode)

    def focus_servo(self, servo_id):
        """Power on designated servo

        Args:
            servo_id: int Joint id 1 - 4
        """
        self.calibration_parameters(class_name=self.__class__.__name__, id=servo_id)
        return self._mesg(ProtocolCode.FOCUS_SERVO, servo_id)

    # Basic for raspberry pi.
    def gpio_init(self):
        """Init GPIO module.
        Raspberry Pi version need this.
        """
        import RPi.GPIO as GPIO  # type: ignore

        GPIO.setmode(GPIO.BCM)
        self.gpio = GPIO

    def gpio_output(self, pin, v):
        """Set GPIO output value.
        Args:
            pin: port number(int).
            v: Output value(int), 1 - GPIO.HEIGH, 0 - GPIO.LOW
        """
        self.gpio.setup(pin, self.gpio.OUT)
        self.gpio.setup(pin, v)

    # Atom IO
    def set_pin_mode(self, pin_no, pin_mode):
        """Set the state mode of the specified pin in atom.

        Args:
            pin_no   (int): pin number.
            pin_mode (int): 0 - input, 1 - output, 2 - input_pullup
        """
        self.calibration_parameters(class_name=self.__class__.__name__, pin_mode=pin_mode)
        return self._mesg(ProtocolCode.SET_PIN_MODE, pin_no, pin_mode)

    def get_gripper_value(self, gripper_type=None):
        """Get the value of gripper.

        Args:
            gripper_type (int): default 1
                1: Adaptive gripper
                3: Parallel gripper
                4: Flexible gripper

        Return:
            gripper value (int)
        """
        if gripper_type is None:
            return self._mesg(ProtocolCode.GET_GRIPPER_VALUE, has_reply=True)
        else:
            self.calibration_parameters(class_name=self.__class__.__name__, gripper_type=gripper_type)
            return self._mesg(ProtocolCode.GET_GRIPPER_VALUE, gripper_type, has_reply=True)

    def is_gripper_moving(self):
        """Judge whether the gripper is moving or not

        Returns:
            0 - not moving
            1 - is moving
            -1- error data
        """
        return self._mesg(ProtocolCode.IS_GRIPPER_MOVING, has_reply=True)

    # Other
    def wait(self, t):
        time.sleep(t)
        return self

    def close(self):
        self.sock.close()



#!/usr/bin/python

from socket import socket, AF_INET, SOCK_STREAM
import sys
import base64
import hashlib
from multiprocessing import Lock
import logging
from pymycobot.log import setup_logging


mutex = Lock()


class ElephantRobot(object):
    def __init__(self, host, port, debug=False):
        # setup connection
        self.debug = debug
        setup_logging(self.debug)
        self.log = logging.getLogger(__name__)
        self.BUFFSIZE = 8 * 1024 * 1024
        self.ADDR = (host, port)
        self.tcp_client = socket(AF_INET, SOCK_STREAM)

    def start_client(self):
        try:
            self.tcp_client.connect(self.ADDR)
            return True
        except Exception as e:
            print(e)
            return False

    def stop_client(self):
        self.tcp_client.close()

    def send_command(self, command):
        with mutex:
            self.tcp_client.send(command.encode())
            recv_data = self.tcp_client.recv(self.BUFFSIZE).decode()
            res_str = str(recv_data)
            if self.debug:
                print("recv = " + res_str)
            res_arr = res_str.split(":")
            if len(res_arr) == 2:
                return res_arr[1]
            else:
                return ""

    def string_to_coords(self, data):
        data = data.replace("[", "")
        data = data.replace("]", "")
        data_arr = data.split(",")
        if len(data_arr) == 6:
            try:
                coords_1 = round(float(data_arr[0]), 3)
                coords_2 = round(float(data_arr[1]), 3)
                coords_3 = round(float(data_arr[2]), 3)
                coords_4 = round(float(data_arr[3]), 3)
                coords_5 = round(float(data_arr[4]), 3)
                coords_6 = round(float(data_arr[5]), 3)
                coords = [coords_1, coords_2, coords_3, coords_4, coords_5, coords_6]
                return coords
            except:
                return self.invalid_coords()
        return self.invalid_coords()

    def string_to_double(self, data):
        try:
            val = round(float(data), 3)
            return val
        except:
            return -9999.99

    def string_to_int(self, data):
        try:
            val = int(data)
            return val
        except:
            return -9999

    def invalid_coords(self):
        coords = [-1.0, -2.0, -3.0, -4.0, -1.0, -1.0]
        return coords

    def get_angles(self):
        command = "get_angles()\n"
        res = self.send_command(command)
        return self.string_to_coords(res)

    def get_coords(self):
        command = "get_coords()\n"
        res = self.send_command(command)
        return self.string_to_coords(res)

    def get_speed(self):
        command = "get_speed()\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def power_on(self):
        command = "power_on()\n"
        res = self.send_command(command)
        return True

    def power_off(self):
        command = "power_off()\n"
        res = self.send_command(command)
        return True

    def start_robot(self):
        command = "start_robot()\n"
        res = self.send_command(command)
        return True

    def check_running(self):
        command = "check_running()\n"
        res = self.send_command(command)
        return res == "1"

    def state_check(self):
        command = "state_check()\n"
        res = self.send_command(command)
        return res == "1"

    def upload_file(self, local_filename, remote_filename):
        """Upload local file to robot via socket. Permissions are checked before
           upload so it cannot overwrite system files (owned by root).

        Args:
            local_filename (str): absolute or relative path to a file to send
            remote_filename (str): absolute or relative file path to where
                                   to upload a file. If path is relative, it is
                                   relative to ~/RoboFlow/upload folder.

        Returns:
            str: ok if success or error message otherwise
        """
        with open(local_filename, "rb") as f:
            content = f.read()
        content_base64 = base64.b64encode(content).decode()
        content_sha256 = hashlib.sha256(content).hexdigest()
        command = "upload_file({},{},{})".format(
            content_base64, remote_filename, content_sha256
        )
        res = self.send_command(command)
        return res

    def read_text_file(self, remote_filename):
        command = f"read_text_file({remote_filename})"
        res = self.send_command(command)
        return res

    def program_open(self, file_path):
        command = "program_open(" + file_path + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def program_run(self, start_line):
        command = "program_run(" + str(start_line) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def read_next_error(self):
        command = "read_next_error()\n"
        res = self.send_command(command)
        return res

    def clear_all_errors(self):
        while self.read_next_error() != "":
            pass

    def write_coords(self, coords, speed):
        command = "set_coords("
        for item in coords:
            command += str(round(item, 3)) + ","
        command += str(speed) + ")\n"
        self.send_command(command)

    def write_coord(self, axis, value, speed):
        coords = self.get_coords()
        if coords != self.invalid_coords():
            coords[axis] = value
            self.write_coords(coords, speed)

    def write_angles(self, angles, speed):
        command = "set_angles("
        for item in angles:
            command += str(round(item, 3)) + ","
        command += str(speed) + ")\n"
        self.send_command(command)

    def write_angle(self, joint, value, speed):
        angles = self.get_angles()
        if angles != self.invalid_coords():
            angles[joint] = value
            self.write_angles(angles, speed)

    def set_speed(self, percentage):
        command = "set_speed(" + str(percentage) + ")\n"
        self.send_command(command)

    def set_carte_torque_limit(self, axis_str, value):
        command = "set_torque_limit(" + axis_str + "," + str(value) + ")\n"
        self.send_command(command)

    def set_upside_down(self, up_down):
        up = "1"
        if up_down:
            up = "0"
        command = "set_upside_down(" + up + ")\n"
        self.send_command(command)

    def set_payload(self, payload):
        command = "set_payload(" + str(payload) + ")\n"
        self.send_command(command)

    def state_on(self):
        command = "state_on()\n"
        self.send_command(command)

    def state_off(self):
        command = "state_off()\n"
        self.send_command(command)

    def task_stop(self):
        command = "task_stop()\n"
        self.send_command(command)

    def jog_angle(self, joint_str, direction, speed):
        command = (
            "jog_angle(" + joint_str + "," + str(direction) + "," + str(speed) + ")\n"
        )
        self.send_command(command)

    def jog_coord(self, axis_str, direction, speed):
        command = (
            "jog_coord(" + axis_str + "," + str(direction) + "," + str(speed) + ")\n"
        )
        self.send_command(command)

    def get_digital_in(self, pin_number):
        command = "get_digital_in(" + str(pin_number) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def get_digital_out(self, pin_number):
        command = "get_digital_out(" + str(pin_number) + ")\n"
        print(command)
        res = self.send_command(command)
        return self.string_to_int(res)

    def get_joint_current(self, joint_number):
        command = "get_joint_current(" + str(joint_number) + ")\n"
        print(command)
        res = self.send_command(command)
        return self.string_to_double(res)

    def set_digital_out(self, pin_number, pin_signal):
        command = "set_digital_out(" + str(pin_number) + "," + str(pin_signal) + ")\n"
        self.send_command(command)

    def set_analog_out(self, pin_number, pin_signal):
        command = "set_analog_out(" + str(pin_number) + "," + str(pin_signal) + ")\n"
        self.send_command(command)

    def send_feed_override(self, override):
        command = "set_feed_rate(" + str(override) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def get_acceleration(self):
        command = "get_acceleration()\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def set_acceleration(self, acceleration):
        command = "set_acceleration(" + str(acceleration) + ")\n"
        self.send_command(command)

    def command_wait_done(self):
        command = "wait_command_done()\n"
        self.send_command(command)

    def wait(self, seconds):
        command = "wait(" + str(seconds) + ")\n"
        self.send_command(command)

    def assign_variable(self, var_name, var_value):
        command = 'assign_variable("' + str(var_name) + '",' + str(var_value) + ")\n"
        self.send_command(command)

    def get_variable(self, var_name):
        command = 'get_variable("' + str(var_name) + '")\n'
        return self.send_command(command)

    def jog_relative(self, joint_id, angle, speed, mode):
        """Relative jog.

        Example:
            jog_relative('J1', 5, 600, 1)

        Args:
            joint_id (str): 'J1' - 'J6', 'X', 'Y', 'Z', 'RX', 'RY', 'RZ'
            angle (float): relative angle or coord value to move, can be negative
                           to move to other direction
            speed (int): speed value
            mode (int): 0 (coord mode) or 1 (joint mode)

        Returns:
            str: return data from socket
        """
        command = "jog_increment({},{},{},{})\n".format(joint_id, angle, speed, mode)
        return self.send_command(command)

    def set_gripper_state(self, state, speed):
        """Sets gripper state.

        Args:
            state (int): gripper state, 0 - open, 1 - close
            speed (int): speed, 1-100

        Returns:
            str: return message
        """
        command = "set_gripper_state(" + str(state) + "," + str(speed) + ")\n"
        return self.send_command(command)

    def set_gripper_value(self, value, speed):
        """Sets gripper open/close state.

        Args:
            value (int): gripper open value, 0-100
            speed (int): speed, 1-100

        Returns:
            str: return message
        """
        command = "set_gripper_value(" + str(value) + "," + str(speed) + ")\n"
        return self.send_command(command)

    def set_gripper_calibrate(self):
        """Sets gripper zero position.

        Returns:
            str: return message
        """
        command = "set_gripper_calibrate()\n"
        return self.send_command(command)

    def set_gripper_enabled(self, enabled):
        """Set gripper enabled.

        Args:
            enabled (int): 1 - enabled, 0 - disabled

        Returns:
            str: return message
        """
        command = "set_gripper_enabled(" + str(enabled) + ")\n"
        return self.send_command(command)

    def set_gripper_mode(self, mode):
        """Sets gripper mode. Default mode is IO, after setting mode 485
           cannot set mode back to IO (need gripper reboot).

        Args:
            mode (int): 0 - 485, 1 - IO

        Returns:
            str: return message
        """
        command = "set_gripper_mode(" + str(mode) + ")\n"
        return self.send_command(command)


if __name__ == "__main__":
    ep = ElephantRobot("192.168.124.28", 5001)
    resp = ep.start_client()
    if resp != "":
        print(resp)
        sys.exit(1)
    print(ep.wait(5))
    print(ep.get_angles())
    print(ep.get_coords())
    print(ep.get_speed())
    # print(ep.power_on())
    # print(ep.power_off())
    print(ep.check_running())
    print(ep.state_check())
    print(ep.program_open("a.tax"))
    print(ep.program_run(0))
    print(ep.read_next_error())
    print(ep.write_coords([1, 2, 3, 4, 5, 6], 110))
    print(ep.write_coord(1, 100, 200))
    print(ep.write_angles([10, 20, 30, 40, 50, 60], 110))
    print(ep.write_angle(3, 180, 200))
    print(ep.set_speed(377))
    print(ep.set_carte_torque_limit("x", 55))
    print(ep.set_upside_down(False))
    print(ep.set_payload(100))
    print(ep.state_on())
    print(ep.state_off())
    print(ep.task_stop())
    print(ep.jog_angle("j2", 1, 300))
    print(ep.jog_coord("rY", 0, 200))
    print(ep.get_digital_in(3))
    print(ep.get_digital_out(3))
    print(ep.set_digital_out(3, 1))
    print(ep.get_digital_out(3))
    print(ep.set_analog_out(1, 3.5))
    print(ep.get_acceleration())
    print(ep.set_acceleration(55))
    print(ep.command_wait_done())
    print(ep.get_variable("f"))
    print(ep.assign_variable("ss", '"eee"'))
    print(ep.get_joint_current(1))

    print(ep.set_gripper_mode(1))
    print(ep.wait(2))
    print(ep.set_gripper_mode(0))

    print(ep.set_gripper_value(100, 20))
    print(ep.wait(2))
    print(ep.set_gripper_value(50, 20))
    print(ep.wait(2))
    print(ep.set_gripper_value(0, 20))
    print(ep.wait(2))

    print(ep.set_gripper_enabled(1))
    print(ep.wait(2))
    print(ep.set_gripper_enabled(0))

    ep.stop_client()

    name = input("input:")
    print(name)




from .protocol_packet_handler import *
# 1 byte
FIRMWARE_MAJOR = 0x00
FIRMWARE_MINOR = 0x01
SERVO_MAJOR = 0x03
SERVO_MINOR = 0x04
BAUD = 6
RETURN_DELAY = 7
PHASE = 18
MAX_TEMPERATURE = 13
MAX_VOLTAGE = 14
SERVO_P = 21
SERVO_D = 22
SERVO_I = 23
CLOCKWISE_INSENSITIVE_ZONE = 26
COUNTERCLOCKWISE_INSENSITIVE_ZONE = 27
D_CONTROL_TIME = 81
# 2 bytes
MIN_START_FORCE = 24


class sms_sts(protocol_packet_handler):
    def __init__(self, portHandler):
        protocol_packet_handler.__init__(self, portHandler, 0)
        
    def get_servo_baud(self, id):
        """获取舵机波特率
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, BAUD)
        return res[0] if res[1] != -2 else None
    
    def set_servo_baud(self, id, value):
        """设置舵机波特率"""
        return self.write1ByteTxRx(id, BAUD, value)
    
    def get_servo_response_speed(self, id):
        """获取舵机响应速度
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, RETURN_DELAY)
        return res[0] if res[1] != -2 else None
    
    def set_servo_response_speed(self, id, value):
        """设置舵机响应速度"""
        return self.write1ByteTxRx(id, RETURN_DELAY, value)
    
    def get_servo_phase(self, id):
        """获取舵机相位
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, PHASE)
        return res[0] if res[1] != -2 else None
    
    def set_servo_phase(self, id, value):
        """设置舵机相位"""
        return self.write1ByteTxRx(id, PHASE, value)
    
    def get_servo_max_temperature(self, id):
        """获取舵机最大温度
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, MAX_TEMPERATURE)
        return res[0] if res[1] != -2 else None
    
    def set_servo_max_temperature(self, id, value):
        """设置舵机最大温度"""
        return self.write1ByteTxRx(id, MAX_TEMPERATURE, value)
       
    def get_servo_max_voltage(self, id):
        """获取舵机最大电压
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, MAX_VOLTAGE)
        return res[0] if res[1] != -2 else None
    
    def set_servo_max_voltage(self, id, value):
        """设置舵机最大电压"""
        return self.write1ByteTxRx(id, MAX_VOLTAGE, value)
    
    def get_servo_pid(self, id):
        """获取舵机pid
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res_p = self.read1ByteTxRx(id, SERVO_P)
        res_i = self.read1ByteTxRx(id, SERVO_I)
        res_d = self.read1ByteTxRx(id, SERVO_D)
        return [res_p[0] if res_p[1] != -2 else None, res_i[0] if res_i[1] != -2 else None, res_d[0] if res_d[1] != -2 else None]
    
    def set_servo_pid(self, id, pid):
        """设置舵机pid"""
        self.write1ByteTxRx(id, SERVO_P, pid[0])
        self.write1ByteTxRx(id, SERVO_I, pid[1])
        self.write1ByteTxRx(id, SERVO_D, pid[2])
    
    def get_servo_clockwise(self, id):
        """获取舵机顺时针不灵敏区
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, CLOCKWISE_INSENSITIVE_ZONE)
        return res[0] if res[1] != -2 else None
    
    def set_servo_clockwise(self, id, value):
        """设置舵机顺时针不灵敏区"""
        return self.write1ByteTxRx(id, CLOCKWISE_INSENSITIVE_ZONE, value)
    
    def get_servo_counter_clockwise(self, id):
        """获取舵机逆时针不灵敏区
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, COUNTERCLOCKWISE_INSENSITIVE_ZONE)
        return res[0] if res[1] != -2 else None
    
    def set_servo_counter_clockwise(self, id, value):
        """设置舵机逆时针不灵敏区"""
        return self.write1ByteTxRx(id, COUNTERCLOCKWISE_INSENSITIVE_ZONE, value)
     
    def get_servo_d_time(self, id):
        """获取舵机D控制时间
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        res = self.read1ByteTxRx(id, D_CONTROL_TIME)
        return res[0] if res[1] != -2 else None
    
    def set_servo_d_time(self, id, value):
        """设置舵机D控制时间"""
        return self.write1ByteTxRx(id, D_CONTROL_TIME, value)
    
    def get_servo_min_start(self, id):
        """获取舵机最小启动力
        
        Args:
            id: 电机Id
            
        Return:
            None: 获取失败
        """
        scs_present_speed, scs_comm_result, scs_error = self.read2ByteTxRx(id, MIN_START_FORCE)
        return self.scs_tohost(scs_present_speed, 15) if scs_comm_result != -2 else None
    
    def set_servo_min_start(self, id, value):
        """设置舵机最小启动力"""
        return self.write2ByteTxRx(id, MIN_START_FORCE, value)
    
    def search_servo(self, id):
        """搜索舵机是否存在
        
        Args:
            id：舵机ID号.
        
        Return:
            0: 存在.
            -6：不存在.
            None: 获取失败.
        """
        res = self.ping(id)
        if res == -2:
            return None
        return res
    
    def get_servo_error(self, id):
        """获取舵机错误信息
        
        Return:
            error_info: 将返回的十进制数据转换为二进制
                Bit0  Bit1  Bit2 Bit3 Bit4 Bit5 对应位被置1表示相应错误出现\n
                电压  传感器 温度 电流 角度  过载 对应位0为无相应该错误\n
                None:表示获取失败
        """
        res = self.read1ByteTxRx(id, 0x41)
        if res[1] == -2:
            return None
        return res[0]
        
    def get_servo_firmware_version(self, id):
        """获取电机固件版本信息
        
        Return:
            list: [固件主版本号, 固件次版本号, 舵机主版本号, 舵机次版本号], -1 表示获取失败
        """
        res = []
        command = [FIRMWARE_MAJOR, FIRMWARE_MINOR, SERVO_MAJOR, SERVO_MINOR]
        for add in command:
            r = self.read1ByteTxRx(id, add)
            if r[1] == -2:
                res.append(-1)
            else:
                res.append(r[0])
        return res
                




# coding=utf-8

import time
import struct
import locale
import numpy as np
from datetime import datetime
import re

from pymycobot.error import calibration_parameters, restrict_serial_port
from pymycobot.common import DataProcessor, ProtocolCode, write, read, ProGripper
from pymycobot.robot_info import _interpret_status_code


class MercuryCommandGenerator(DataProcessor):
    _write = write
    _read = read

    def __init__(self, debug=False):
        super(MercuryCommandGenerator, self).__init__(debug)
        self.calibration_parameters = calibration_parameters
        # self.is_stop = False
        self.write_command = []
        self.read_command = []
        # 同步模式
        self.sync_mode = True
        self.language, _ = locale.getdefaultlocale()
        if self.language not in ["zh_CN", "en_US"]:
            self.language = "en_US"
        self.max_joint, self.min_joint = 0,0
        self.all_debug_data = []
        self.all_read_data = b""
            
    def _joint_limit_init(self):
        max_joint = np.zeros(7)
        min_joint = np.zeros(7)
        for i in range(7):
            max_joint[i] = self.get_joint_max_angle(i + 1)
            min_joint[i] = self.get_joint_min_angle(i + 1)
        return max_joint, min_joint
    
    def _joint_limit_judge(self, angles):
        offset = 3 
        try:
            for i in range(7):
                if self.min_joint[i] + offset < angles[i] < self.max_joint[i] - offset:
                    pass
                else:
                    if self.language == "zh_CN":
                        return f"当前角度为{angles[i]}, 角度范围为： {self.min_joint[i]} ~ {self.max_joint[i]}"
                    return f"current value = {angles[i]}, limit is {self.min_joint[i]} ~ {self.max_joint[i]}"
        except TypeError:
            return "joint limit error"
        return "over limit error {}".format(angles)
            
    def _Singularity(self, angles):
        try:
            singular_angles = [0, 180]  # Joint 6: 0 and 180 degrees are singular points
            state = ""
            offset = 5
            for singular in singular_angles:
                if singular - offset < angles[5] < singular + offset:
                    if self.language == "zh_CN":
                        return f"在关节 6 处检测到奇点：{angles[5]} 度"
                    return f"Singularity detected at joint 6: {angles[5]} degrees"
            return state
        except:
            return "Singularity error"

    def _check_coords(self, new_coords, is_print=0):
        try:
            first_three = new_coords[:3]
            first_three[2] -= 83.64
            info = ""
            # Calculate the Euclidean norm (magnitude)
            magnitude = np.linalg.norm(first_three)
            if is_print == 1:
                if self.language == "zh_CN":
                    info +=f"当前臂展为{magnitude}, 最大的臂展为{self.arm_span}"
                else:
                    info +=f"Arm span is {magnitude}, max is {self.arm_span}"

            # if magnitude > self.arm_span - 10:
            #     if self.language == "zh_CN":
            #         info += f"当前臂展为{magnitude}超出物理限位, 最大的臂展为{self.arm_span}"
            #     else:
            #         info += f"Arm span is {magnitude} exceeds physical limit, max is {self.arm_span}"
            return info
        except:
            return "check coords error"

    def _status_explain(self, status):
        error_info = _interpret_status_code(self.language, status)
        if error_info != "":
            self.arm_span = 440
        if 0x00 < status <= 0x07:
            angles = self.get_angles()
            if type(self.max_joint) == int and self.max_joint == 0:
                self.max_joint, self.min_joint = self._joint_limit_init()
            error_info+=self._joint_limit_judge(angles)
        elif status in [32,33]:
            error_coords = self.get_coords()
            error_info += self._check_coords(error_coords, 1)
        elif status == 36:
            angles = self.get_angles()
            error_info += self._Singularity(angles)
        
        return error_info
            
    def _send_command(self, genre, real_command):
        self.write_command.append(genre)
        if self.__class__.__name__ == "Mercury":
            self._write(self._flatten(real_command))
        elif self.__class__.__name__ == "MercurySocket":
            self._write(self._flatten(real_command), method="socket")

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply, _async = super(
            MercuryCommandGenerator, self)._mesg(genre, *args, **kwargs)
        is_in_position = False
        is_get_return = False
        lost_times = 0
        with self.lock:
            self._send_command(genre, real_command)
        if _async:
            with self.lock:
                if genre in self.write_command:
                    self.write_command.remove(genre)
                return None
        t = time.time()
        wait_time = 0.15
        big_wait_time = False
        if genre == ProtocolCode.POWER_ON:
            wait_time = 8
            big_wait_time = True
        elif genre in [ProtocolCode.POWER_OFF, ProtocolCode.POWER_ON_ONLY,  ProtocolCode.RELEASE_ALL_SERVOS, ProtocolCode.FOCUS_ALL_SERVOS,
                    ProtocolCode.RELEASE_SERVO, ProtocolCode.FOCUS_SERVO, ProtocolCode.STOP, ProtocolCode.SET_CONTROL_MODE, ProtocolCode.MERCURY_DRAG_TEACH_CLEAN]:
            wait_time = 3
            big_wait_time = True
        elif genre in [
                ProtocolCode.SEND_ANGLE,
                ProtocolCode.SEND_ANGLES,
                ProtocolCode.SEND_COORD,
                ProtocolCode.SEND_COORDS,
                ProtocolCode.JOG_ANGLE,
                ProtocolCode.JOG_COORD,
                ProtocolCode.JOG_INCREMENT,
                ProtocolCode.JOG_INCREMENT_COORD,
                ProtocolCode.COBOTX_SET_SOLUTION_ANGLES,
                ProtocolCode.MERCURY_SET_BASE_COORDS,
                ProtocolCode.MERCURY_JOG_BASE_COORD,
                ProtocolCode.MERCURY_SET_BASE_COORD,
                ProtocolCode.OVER_LIMIT_RETURN_ZERO,
                ProtocolCode.JOG_BASE_INCREMENT_COORD,
                ProtocolCode.WRITE_MOVE_C,
                ProtocolCode.JOG_RPY,
                ProtocolCode.WRITE_MOVE_C_R] and self.sync_mode:
            wait_time = 300
            is_in_position = True
            big_wait_time = True
        elif genre in [ProtocolCode.SERVO_RESTORE]:
            wait_time = 0.3
            big_wait_time = True
        need_break = False
        data = None
        timeout = 0.5
        if self.__class__.__name__ == "MercurySocket":
            timeout = 1
        interval_time = time.time()
        is_moving = 0
        check_is_moving_t = 1
        while True and time.time() - t < wait_time:
            with self.lock_out:
                for v in self.read_command:
                    read_data = v[0]
                    if is_get_return and is_in_position and read_data[2] == 0x04 and read_data[3] == 0x5b:
                        if v[1] < t:
                            with self.lock:
                                if v in self.read_command:
                                    self.read_command.remove(v)
                            continue
                        # print("到位反馈", flush=True)
                        is_get_return = True
                        need_break = True
                        data = read_data
                        with self.lock:
                            if v in self.read_command:
                                self.read_command.remove(v)
                            if genre in self.write_command:
                                self.write_command.remove(genre)

                    elif genre == read_data[3] and read_data[2] == 5 and read_data[4] == 0xFF:
                        # 通信闭环
                        # print(-2)
                        # print("闭环", flush=True)
                        is_get_return = True
                        with self.lock:
                            if v in self.read_command:
                                self.read_command.remove(v)
                        if has_reply == False or self.sync_mode == False:
                            # print(-3)
                            # print("仅闭环退出", flush=True)
                            need_break = True
                            data = read_data
                    elif genre == read_data[3]:
                        # print(-4)
                        # print("正常读取", flush=True)
                        need_break = True
                        data = read_data
                        with self.lock:
                            if v in self.read_command:
                                self.read_command.remove(v)
                            if genre in self.write_command:
                                self.write_command.remove(genre)
                        break
            if (not big_wait_time or is_in_position) and not is_get_return and time.time() - t > timeout:
                t = time.time()
                moving = self.is_moving()
                if isinstance(moving, int):
                    if moving != 0:
                        continue
                # 运动指令丢失，重发
                # print("运动指令丢失，重发", flush=True)
                lost_times += 1
                # print("运动指令丢失，重发")
                with self.lock:
                    self._send_command(genre, real_command)
            if need_break:
                # print("正常退出", flush=True)
                break
            if lost_times > 2:
                # 重传3次失败，返回-1
                # print("重传3次失败，返回-1", flush=True)
                return -1
            # if t < self.is_stop and genre != ProtocolCode.STOP:
            #     # 打断除了stop指令外的其他指令的等待
            #     self.is_stop = 0
            #     break
            if is_in_position and time.time() - interval_time > check_is_moving_t and wait_time == 300:
                interval_time = time.time()
                moving = self.is_moving()
                if isinstance(moving, int) and moving == 0:
                    # print("停止运动，退出")
                    is_moving += 1
                    check_is_moving_t = 0.25 # 由于is_moving接口比到位反馈更快，所以第一次收到停止运动后，将下一次的检测时间更改为0.25s，防止此处先退出，返回-2
                    if is_moving > 1:
                        # 累计两次才退出
                        with self.lock:
                            if genre in self.write_command:
                                self.write_command.remove(genre)
                        return -2
            time.sleep(0.001)
        else:
            # print("ERROR: ---超时---"
            pass
        if data is None:
            # print("未拿到数据")
            return data
        res = []
        data = bytearray(data)
        data_len = data[2] - 3
        # unique_data = [ProtocolCode.GET_BASIC_INPUT,
        #                ProtocolCode.GET_DIGITAL_INPUT]
        if genre == ProtocolCode.GET_BASIC_INPUT:
            data_pos = 5
            data_len -= 1
        else:
            data_pos = 4
        if is_get_return:
            data_len -= 1
            data_pos += 1
            if data[2] == 5:
                # print("握手成功")
                return data[5]
            elif data[2] == 4:
                res= self._status_explain(data[4])
                if res != "":
                    print(res)
                return data[4]
        valid_data = data[data_pos: data_pos + data_len]
        # print(data_len, valid_data)
        if data_len in [6, 8, 12, 14, 16, 20, 24, 26, 60]:
            if data_len == 8 and (genre == ProtocolCode.IS_INIT_CALIBRATION):
                if valid_data[0] == 1:
                    return 1
                n = len(valid_data)
                for v in range(1, n):
                    res.append(valid_data[v])
            elif data_len == 8 and genre == ProtocolCode.GET_DOWN_ENCODERS:
                i = 0
                while i < data_len:
                    byte_value = int.from_bytes(
                        valid_data[i:i+4], byteorder='big', signed=True)
                    i += 4
                    res.append(byte_value)
            elif data_len == 6 and genre in [ProtocolCode.GET_SERVO_STATUS, ProtocolCode.GET_SERVO_VOLTAGES, ProtocolCode.GET_SERVO_CURRENTS]:
                for i in range(data_len):
                    res.append(valid_data[i])
            else:
                for header_i in range(0, len(valid_data), 2):
                    one = valid_data[header_i: header_i + 2]
                    res.append(self._decode_int16(one))
        elif data_len == 2:
            if genre in [ProtocolCode.IS_SERVO_ENABLE]:
                return self._decode_int8(valid_data[1:2])
            elif genre in [ProtocolCode.GET_LIMIT_SWITCH]:
                for i in valid_data:
                    res.append(i)
            elif genre in [ProtocolCode.GET_ERROR_INFO]:
                res.append(self._decode_int8(valid_data[1:]))
            else:
                res.append(self._decode_int16(valid_data))
        elif data_len == 3:
            if genre in [ProtocolCode.GET_DIGITAL_INPUTS]:
                for i in valid_data:
                    res.append(i)
            else:
                res.append(self._decode_int16(valid_data[1:]))
        elif data_len == 4:
            if genre in [ProtocolCode.COBOTX_GET_ANGLE, ProtocolCode.GET_ENCODER, ProtocolCode.GET_DYNAMIC_PARAMETERS]:
                byte_value = int.from_bytes(
                    valid_data, byteorder='big', signed=True)
                if genre in [ProtocolCode.GET_DYNAMIC_PARAMETERS]:
                    byte_value /= 1000
                res.append(byte_value)
            else:
                for i in range(1, 4):
                    res.append(valid_data[i])
        elif data_len == 7:
            # if genre ==ProtocolCode.GET_TORQUE_COMP:
            for i in range(0, data_len):
                res.append(valid_data[i])
            # else:
            #     error_list = [i for i in valid_data]
            #     for i in error_list:
            #         if i in range(16, 23):
            #             res.append(1)
            #         elif i in range(23, 29):
            #             res.append(2)
            #         elif i in range(32, 112):
            #             res.append(3)
            #         else:
            #             res.append(i)
        elif data_len in [28, 32]: # 28 left get_zero_pos, 32 right get_zero_pos
            for i in range(0, data_len, 4):
                byte_value = int.from_bytes(
                    valid_data[i:i+4], byteorder='big', signed=True)
                res.append(byte_value)
        elif data_len == 40 and genre == ProtocolCode.GET_ANGLES_COORDS:
            i = 0
            while i < data_len:
                if i < 28:
                    byte_value = int.from_bytes(
                        valid_data[i:i+4], byteorder='big', signed=True)
                    res.append(byte_value)
                    i += 4
                else:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 40 and genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            # 右臂上位机错误
            i = 0
            res = []
            while i < data_len:
                if i < 10 or i >= 30:
                    res.append(valid_data[i])
                    i += 1
                elif i < 30:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 41 and genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            # 图灵右臂上位机错误
            i = 0
            res = []
            while i < data_len:
                if i < 9:
                    res.append(valid_data[i])
                    i += 1
                else:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len in [37, 46]:
            # X1 left status
            if data_len == 37:
                add_index = 9
            else:
                add_index = 10
            i = 0
            res = []
            while i < data_len:
                if i < add_index:
                    res.append(valid_data[i])
                    i += 1
                else:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 38:
            i = 0
            res = []
            while i < data_len:
                if i < 10 or i >= 30:
                    res.append(valid_data[i])
                    i += 1
                elif i < 38:
                    one = valid_data[i: i + 2]
                    res.append(self._decode_int16(one))
                    i += 2
        elif data_len == 56:
            i = 0
            while i < data_len:
                byte_value_send = int.from_bytes(valid_data[i:i+4], byteorder='big', signed=True)
                i+=4
                byte_value_current = int.from_bytes(valid_data[i:i+4], byteorder='big', signed=True)
                res.append([byte_value_send, byte_value_current])
                i+=4
        elif data_len == 48:
            i = 0
            j = 0
            res = [0, 0]
            angles = []
            coords = []
            while i < data_len:
                if i < 28:
                    byte_value_send = int.from_bytes(valid_data[i:i+4], byteorder='big', signed=True)
                    angles.append(self._int2angle(byte_value_send))
                    i+=4
                elif i < 40:
                    
                    one = valid_data[i: i + 2]
                    one = self._decode_int16(one)
                    if j < 3:
                        one = self._int2coord(one)
                    else:
                        one = self._int2angle(one)
                    coords.append(one)
                    j += 1
                    i += 2
                else:
                    res.append(valid_data[i])
                    i += 1
            res[0] = angles
            res[1] = coords
            res[2] = 0
            res[3] = 0
        elif genre == ProtocolCode.MERCURY_GET_TOQUE_GRIPPER:
            res = self._decode_int16(valid_data[-2:])
            address = self._decode_int16(valid_data[1:3])
            if address == 1:
                res /= 10
        else:
            if genre in [
                ProtocolCode.GET_SERVO_VOLTAGES,
                ProtocolCode.GET_SERVO_STATUS,
                ProtocolCode.GET_SERVO_TEMPS,
            ]:
                for i in range(data_len):
                    data1 = self._decode_int8(valid_data[i: i + 1])
                    res.append(0xFF & data1 if data1 < 0 else data1)
            res.append(self._decode_int8(valid_data))
        if res == []:
            return None
        if genre in [
            ProtocolCode.ROBOT_VERSION,
            ProtocolCode.GET_ROBOT_ID,
            ProtocolCode.IS_POWER_ON,
            ProtocolCode.IS_CONTROLLER_CONNECTED,
            ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
            ProtocolCode.IS_IN_POSITION,
            ProtocolCode.IS_MOVING,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.IS_ALL_SERVO_ENABLE,
            ProtocolCode.GET_SERVO_DATA,
            ProtocolCode.GET_DIGITAL_INPUT,
            ProtocolCode.GET_GRIPPER_VALUE,
            ProtocolCode.IS_GRIPPER_MOVING,
            ProtocolCode.GET_SPEED,
            ProtocolCode.GET_ENCODER,
            ProtocolCode.GET_BASIC_INPUT,
            ProtocolCode.GET_TOF_DISTANCE,
            ProtocolCode.GET_END_TYPE,
            ProtocolCode.GET_MOVEMENT_TYPE,
            ProtocolCode.GET_REFERENCE_FRAME,
            ProtocolCode.GET_FRESH_MODE,
            ProtocolCode.GET_GRIPPER_MODE,
            ProtocolCode.SET_SSID_PWD,
            # ProtocolCode.GET_ERROR_DETECT_MODE,
            ProtocolCode.POWER_ON,
            ProtocolCode.POWER_OFF,
            ProtocolCode.RELEASE_ALL_SERVOS,
            ProtocolCode.RELEASE_SERVO,
            ProtocolCode.FOCUS_ALL_SERVOS,
            ProtocolCode.FOCUS_SERVO,
            ProtocolCode.STOP,
            ProtocolCode.SET_BREAK,
            ProtocolCode.IS_BTN_CLICKED,
            ProtocolCode.GET_CONTROL_MODE,
            ProtocolCode.GET_VR_MODE,
            ProtocolCode.GET_FILTER_LEN,
            ProtocolCode.MERCURY_ERROR_COUNTS,
            ProtocolCode.GET_MAX_ACC,
            ProtocolCode.GET_MONITOR_MODE,
            ProtocolCode.GET_COLLISION_MODE,
            ProtocolCode.GET_DYNAMIC_PARAMETERS,
            ProtocolCode.GET_ERROR_INFO
        ]:
            return self._process_single(res)
        elif genre in [ProtocolCode.GET_DRAG_FIFO]:
            return [self._int2angle(angle) for angle in res]
        elif genre in [
            ProtocolCode.GET_COORDS,
            ProtocolCode.MERCURY_GET_BASE_COORDS,
            ProtocolCode.GET_TOOL_REFERENCE,
            ProtocolCode.GET_WORLD_REFERENCE,
        ]:
            if res:
                r = []
                for idx in range(3):
                    r.append(self._int2coord(res[idx]))
                for idx in range(3, 6):
                    r.append(self._int2angle(res[idx]))
                return r
            else:
                return res
        elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
            return [self._int2coord(angle) for angle in res]
        elif genre in [ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION]:
            return self._int2coord(self._process_single(res))
        elif genre in [
            ProtocolCode.GET_JOINT_MAX_ANGLE,
            ProtocolCode.GET_JOINT_MIN_ANGLE,
        ]:
            return self._int2coord(res[0])
        elif genre == ProtocolCode.GET_ANGLES_COORDS:
            r = []
            for index in range(len(res)):
                if index < 7:
                    r.append(self._int2angle(res[index]))
                elif index < 10:
                    r.append(self._int2coord(res[index]))
                else:
                    r.append(self._int2angle(res[index]))
            return r
        elif genre == ProtocolCode.GO_ZERO:
            r = []
            if res:
                if 1 not in res[1:]:
                    return res[0]
                else:
                    for i in range(1, len(res)):
                        if res[i] == 1:
                            r.append(i)
            return r
        elif genre in [ProtocolCode.COBOTX_GET_SOLUTION_ANGLES, ProtocolCode.MERCURY_GET_POS_OVER_SHOOT, ProtocolCode.GET_SERVO_CW]:
            return self._int2angle(res[0])
        elif genre == ProtocolCode.GET_ANGLES:
            return [self._int3angle(angle) for angle in res]
        elif genre == ProtocolCode.SOLVE_INV_KINEMATICS:
            return [self._int2angle(angle) for angle in res]
        elif genre == ProtocolCode.COBOTX_GET_ANGLE:
            return self._int2angle(res[0])
        elif genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            if len(res) == 23:
                index = 9
            else:
                index = 10
            for i in range(index, len(res)):
                if res[i] != 0:
                    data = bin(res[i])[2:]
                    res[i] = []
                    while len(data) != 16:
                        data = "0"+data
                    for j in range(16):
                        if data[j] != "0":
                            error_id = 15-j
                            res[i].append(error_id)
                    if res[i] == []:
                        res[i] = 0
            return res
        elif genre == ProtocolCode.GET_MOTORS_RUN_ERR:
            for i in range(len(res)):
                if res[i] != 0:
                    data = bin(res[i])[2:]
                    res[i] = []
                    while len(data) != 16:
                        data = "0"+data
                    for j in range(16):
                        if data[j] != "0":
                            error_id = 15-j
                            res[i].append(error_id)
                    if res[i] == []:
                        res[i] = 0
            return res
        else:
            if isinstance(res, list) and len(res) == 1:
                return res[0]
            return res

    def _process_received(self, data):
        if not data:
            return []

        data = bytearray(data)
        data_len = len(data)
        # Get valid header: 0xfe0xfe
        header_i, header_j = 0, 1
        while header_j < data_len - 4:
            if self._is_frame_header(data, header_i, header_j):
                cmd_id = data[header_i + 3]
                if cmd_id in self.write_command or cmd_id == 0x5B:
                    return data
            header_i += 1
            header_j += 1
        else:
            return []

    def read_thread(self, method=None):
        all_data = b''
        while True:
            try:
                datas = b""
                data_len = -1
                k = 0
                pre = 0
                t = time.time()
                wait_time = 0.5
                if method is not None:
                    try:
                        self.sock.settimeout(wait_time)
                        data = self.sock.recv(1024)
                        if isinstance(data, str):
                            datas = bytearray()
                            for i in data:
                                datas += hex(ord(i))
                            data = datas
                    except:
                        data = b""
                    if data != b"":
                        pattern = re.compile(rb'###(.*?)###', re.DOTALL)
                        matches = pattern.findall(data)
                        for match in matches:
                            if self.check_python_version() == 2:
                                command_log = ""
                                for d in match:
                                    command_log += hex(ord(d))[2:] + " "
                                self.log.debug("_read : {}".format(command_log))
                                # self.log.debug("_read: {}".format([hex(ord(d)) for d in data]))
                            else:
                                command_log = ""
                                for d in match:
                                    command_log += hex(d)[2:] + " "
                                self.log.debug("_read : {}".format(command_log))
                            res = self._process_received(match)
                            if res != []:
                                with self.lock:
                                    self.read_command.append([res, time.time()])
                else:
                    while True and time.time() - t < wait_time:
                        if self._serial_port.isOpen() and self._serial_port.inWaiting() > 0:
                            data = self._serial_port.read()
                            if self.save_serial_log:
                                all_data+=data
                            # self.log.info(all_read_data)
                            k += 1
                            # print(datas, flush=True)
                            if data_len == 3:
                                datas += data
                                crc = self._serial_port.read(2)
                                self.all_read_data += data
                                if self.crc_check(datas) == [v for v in crc]:
                                    datas += crc
                                    break
                            if data_len == 1 and data == b"\xfa":
                                datas += data
                                # if [i for i in datas] == command:
                                #     datas = b''
                                #     data_len = -1
                                #     k = 0
                                #     pre = 0
                                #     continue
                                # break
                            elif len(datas) == 2:
                                data_len = struct.unpack("b", data)[0]
                                datas += data
                            elif len(datas) > 2 and data_len > 0:
                                datas += data
                                data_len -= 1
                            elif data == b"\xfe":
                                if datas == b"":
                                    datas += data
                                    pre = k
                                else:
                                    if k - 1 == pre:
                                        datas += data
                                    else:
                                        datas = b"\xfe"
                                        pre = k
                        else:
                            time.sleep(0.001)
                    else:
                        datas = b''
                    if self.save_serial_log:
                        if all_data != b'':
                            current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                            with open("all_log.txt", "a") as f:
                                f.write(str(current_time)+str(all_data)[2:-1]+"\n")
                            all_data = b''
                    if datas != b'':
                        res = self._process_received(datas)
                        if self.check_python_version() == 2:
                            command_log = ""
                            for d in datas:
                                data = hex(d)[2:]
                                if len(data) != 2:
                                    data = "0" + data
                                command_log += data + " "
                            self.log.debug("_read : {}".format(command_log))
                        else:
                            command_log = ""
                            for d in datas:
                                data = hex(d)[2:]
                                if len(data) != 2:
                                    data = "0" + data
                                command_log += data + " "
                            self.log.debug("_read : {}".format(command_log))
                        if datas[3] == 0x5D:
                            debug_data = []
                            for i in range(4, 32, 4):
                                byte_value = int.from_bytes(datas[i:i+4], byteorder='big', signed=True)
                                debug_data.append(byte_value)
                            # self.log.info("SERVO_COMMAND : {}".format(debug_data))
                            with open("plotPos.txt", "a") as f:
                                f.write(str(debug_data)+"\n")
                            continue
                        elif datas[3] == 0x8E and datas[2] == 0x3B:
                            debug_data = []
                            
                            for i in range(4, 60, 2):
                                if i < 40:
                                    data = self._decode_int16(datas[i:i+2]) / 1000
                                else:
                                    data = self._decode_int16(datas[i:i+2])
                                debug_data.append(data)
                            self.all_debug_data.append(debug_data)
                            
                            continue
                        if res == []:
                            # print("res is empty")
                            continue
                        # if datas[3] == 0x5b:
                        #     print("等待加入到读取队列")
                        with self.lock:
                            self.read_command.append([res, time.time()])
                            # if datas[3] == 0x5b:
                                # print("加入到读取队列成功")
                            
            except Exception as e:
                # self.log.error("read error: {}".format(traceback.format_exc()))
                pass
            time.sleep(0.001)

    def get_system_version(self):
        """get system version"""
        return self._mesg(ProtocolCode.SOFTWARE_VERSION)

    def get_atom_version(self):
        """Get atom firmware version.

        Returns:
            float: version number.
        """
        return self._mesg(ProtocolCode.GET_ATOM_VERSION)

    def is_power_on(self):
        """Adjust robot arm status

        Return:
            1 - power on
            0 - power off
            -1 - error data
        """
        return self._mesg(ProtocolCode.IS_POWER_ON)

    def get_coords(self, angles=None):
        """Get the coords from robot arm, coordinate system based on base. The target angle can be passed in, and the coordinate position corresponding to the target angle can be obtained.
        
        Args:
            angles (list): The angles of six joints. e.g. [0, 0, 0, 0, 0, 0]

        Return:
            list : A float list of coord . [x, y, z, rx, ry, rz].

        """
        if angles is None:
            return self._mesg(ProtocolCode.GET_COORDS)
        self.calibration_parameters(
            class_name=self.__class__.__name__, angles=angles)
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.GET_COORDS, angles)

    def is_paused(self):
        """Judge whether the manipulator pauses or not.

        Return:
            1 - paused
            0 - not paused
            -1 - error
        """
        return self._mesg(ProtocolCode.IS_PAUSED)

    def set_solution_angles(self, angle, speed):
        """Set zero space deflection angle value

        Args:
            angle: Angle of joint 1. The angle range is -90 ~ 90
            speed: 1 - 100.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, speed=speed, solution_angle=angle
        )
        return self._mesg(
            ProtocolCode.COBOTX_SET_SOLUTION_ANGLES, [self._angle2int(angle)], speed, has_reply=True
        )

    def get_solution_angles(self):
        """Get zero space deflection angle value"""
        return self._mesg(ProtocolCode.COBOTX_GET_SOLUTION_ANGLES)

    def write_move_c(self, transpoint, endpoint, speed):
        """Circular trajectory motion

        Args:
            transpoint (list): Arc passing point coordinates
            endpoint (list): Arc end point coordinates
            speed (int): 1 ~ 100
        """
        start = []
        end = []
        for index in range(6):
            if index < 3:
                start.append(self._coord2int(transpoint[index]))
                end.append(self._coord2int(endpoint[index]))
            else:
                start.append(self._angle2int(transpoint[index]))
                end.append(self._angle2int(endpoint[index]))
        return self._mesg(ProtocolCode.WRITE_MOVE_C, start, end, speed, has_reply=True)

    def focus_all_servos(self):
        """Lock all joints"""
        return self._mesg(ProtocolCode.FOCUS_ALL_SERVOS)

    def go_home(self, robot=1, speed=20, _async=False):
        """Control the machine to return to the zero position.

        Args:
            robot (int): 
                1 - left arm
                2 - right arm
            speed (int): 1 ~ 100
        Return:
            1 : All motors return to zero position.
            0 : failed.
        """
        if robot == 1:
            return self.send_angles([0, 0, 0, 0, 0, 90, 0], speed, _async=_async)
        else:
            self.send_angle(11, 0, speed)
            self.send_angle(12, 0, speed)
            self.send_angle(13, 0, speed)
            return self.send_angles([0, 0, 0, 0, 0, 90, 0], speed, _async=_async)

    def get_angle(self, joint_id):
        """Get single joint angle

        Args:
            joint_id (int): 1 ~ 7 or 11 ~ 13.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.COBOTX_GET_ANGLE, joint_id)

    def get_angles(self):
        """ Get the angle of all joints.

        Return:
            list: A float list of all angle.
        """
        return self._mesg(ProtocolCode.GET_ANGLES)
    

    def servo_restore(self, joint_id):
        """Abnormal recovery of joints

        Args:
            joint_id (int): Joint ID.
                arm : 1 ~ 7 
                waist : 13
                All joints: 254
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, servo_restore=joint_id
        )
        return self._mesg(ProtocolCode.SERVO_RESTORE, joint_id)

    # def set_error_detect_mode(self, mode):
    #     """Set error detection mode. Turn off without saving, default to open state

    #     Return:
    #         mode : 0 - close 1 - open.
    #     """
    #     self.calibration_parameters(
    #         class_name=self.__class__.__name__, mode=mode
    #     )
    #     self._mesg(ProtocolCode.SET_ERROR_DETECT_MODE, mode)

    # def get_error_detect_mode(self):
    #     """Set error detection mode"""
    #     return self._mesg(ProtocolCode.GET_ERROR_DETECT_MODE, has_reply=True)

    def sync_send_angles(self, degrees, speed, timeout=15):
        """Send the angle in synchronous state and return when the target point is reached

        Args:
            degrees: a list of degree values(List[float]), length 6.
            speed: (int) 0 ~ 100
            timeout: default 7s.
        """
        t = time.time()
        self.send_angles(degrees, speed)
        while time.time() - t < timeout:
            f = self.is_in_position(degrees, 0)
            if f == 1:
                return 1
            time.sleep(0.1)
        return 0

    @restrict_serial_port
    def get_base_coords(self):
        """get base coords"""
        return self._mesg(ProtocolCode.MERCURY_GET_BASE_COORDS)

    @restrict_serial_port
    def send_base_coord(self, axis, coord, speed):
        """Single coordinate control with the torso base as the coordinate system

        Args:
            axis (int): 1 to 6 correspond to x, y, z, rx, ry, rz
            coord (float): coord value.
                The coord range of `X` is -351.11 ~ 566.92.
                The coord range of `Y` is -645.91 ~ 272.12.
                The coord range of `Y` is -262.91 ~ 655.13.
                The coord range of `RX` is -180 ~ 180.
                The coord range of `RY` is -180 ~ 180.
                The coord range of `RZ` is -180 ~ 180.
            speed (int): 1 ~ 100
        """
        if axis < 4:
            coord = self._coord2int(coord)
        else:
            coord = self._angle2int(coord)
        return self._mesg(ProtocolCode.MERCURY_SET_BASE_COORD, axis, [coord], speed)

    @restrict_serial_port
    def send_base_coords(self, coords, speed):
        """Full coordinate control

        Args:
            coords (list): coordinate value, [x, y, z, rx, ry, rz]
            speed (int): 1 ~ 100
        """
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.MERCURY_SET_BASE_COORDS, coord_list, speed)

    @restrict_serial_port
    def jog_base_coord(self, axis, direction, speed, _async=True):
        """Single-coordinate unidirectional motion control

        Args:
            axis (int): 1 ~ 6 correspond to [x, y, z, rx, ry, rz]
            direction (int): Direction of movement. 0 or 1.
            speed (int): 1 ~ 100.
            sync (bool): Whether to wait for the movement to complete. Default to True.
            
        Return:
            0: End of exercise.
            1 ~ 7: Corresponding joint exceeds limit.
            16 ~ 19: Collision protection.
            32 ~ 35: Coordinates have no adjacent solutions.
            49: Not powered on.
            50: Motor abnormality.
            51: Motor encoder error
            52: Not reaching the designated location or not reaching the designated location for more than 5 minutes (only J11, J12 available)
        """
        return self._mesg(ProtocolCode.MERCURY_JOG_BASE_COORD, axis, direction, speed, _async=_async, has_reply=True)

    def drag_teach_save(self):
        """Start recording the dragging teaching point. In order to show the best sports effect, the recording time should not exceed 90 seconds."""
        return self._mesg(ProtocolCode.MERCURY_DRAG_TECH_SAVE)

    def drag_teach_execute(self):
        """Start dragging the teaching point and only execute it once."""
        return self._mesg(ProtocolCode.MERCURY_DRAG_TECH_EXECUTE)

    def drag_teach_pause(self):
        """Pause recording of dragging teaching point"""
        return self._mesg(ProtocolCode.MERCURY_DRAG_TECH_PAUSE)

    # def is_gripper_moving(self, mode=None):
    #     """Judge whether the gripper is moving or not

    #     Args:
    #         mode: 1 - pro gripper(default)  2 - Parallel gripper

    #     Returns:
    #         0 - not moving
    #         1 - is moving
    #         -1- error data
    #     """
    #     if mode:
    #         return self._mesg(ProtocolCode.IS_GRIPPER_MOVING, mode, has_reply=True)
    #     return self._mesg(ProtocolCode.IS_GRIPPER_MOVING, has_reply=True)

    def set_gripper_enabled(self, value):
        """Pro adaptive gripper enable setting

        Args:
            value (int): 
                1 : enable
                0 : release
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, value=value)
        return self._mesg(ProtocolCode.SET_GRIPPER_ENABLED, value)

    def is_btn_clicked(self):
        """Check if the end button has been pressed.

        Return:
            1 : pressed.
            0 : not pressed.
        """
        return self._mesg(ProtocolCode.IS_BTN_CLICKED)

    # def tool_serial_restore(self):
    #     """485 factory reset
    #     """
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_RESTORE)

    # def tool_serial_ready(self):
    #     """Set up 485 communication

    #     Return:
    #         0 : not set
    #         1 : Setup completed
    #     """
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_READY, has_reply=True)

    # def tool_serial_available(self):
    #     """Read 485 buffer length

    #     Return:
    #         485 buffer length available for reading
    #     """
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_AVAILABLE, has_reply=True)

    # def tool_serial_read_data(self, data_len):
    #     """Read fixed length data. Before reading, read the buffer length first. After reading, the data will be cleared

    #     Args:
    #         data_len (int): The number of bytes to be read, range 1 ~ 45
    #     """
    #     self.calibration_parameters(
    #         class_name=self.__class__.__name__, data_len=data_len)
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_READ_DATA, data_len, has_reply=True)

    # def tool_serial_write_data(self, command):
    #     """End 485 sends data， Data length range is 1 ~ 45 bytes

    #     Args:
    #         command : data instructions

    #     Return:
    #         number of bytes received
    #     """
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_WRITE_DATA, command, has_reply=True)

    # def tool_serial_flush(self):
    #     """Clear 485 buffer
    #     """
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_FLUSH)

    # def tool_serial_peek(self):
    #     """View the first data in the buffer, the data will not be cleared

    #     Return:
    #         1 byte data
    #     """
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_PEEK, has_reply=True)

    # def tool_serial_set_baud(self, baud=115200):
    #     """Set 485 baud rate, default 115200

    #     Args:
    #         baud (int): baud rate
    #     """
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_SET_BAUD, baud)

    # def tool_serial_set_timeout(self, max_time):
    #     """Set 485 timeout in milliseconds, default 30ms

    #     Args:
    #         max_time (int): timeout
    #     """
    #     self.calibration_parameters(
    #         class_name=self.__class__.__name__, max_time=max_time)
    #     return self._mesg(ProtocolCode.TOOL_SERIAL_SET_TIME_OUT, max_time)

    def get_robot_status(self):
        return self._mesg(ProtocolCode.MERCURY_ROBOT_STATUS)

    def power_on(self, mode=1):
        """Power on the robot
        
        Args:
            mode (int, optional): 1 - Motor enable, 0 - The motor is not enabled, only initialized.
        
        Return:
            1: success
            2: failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, mode=mode)
        res = self._mesg(ProtocolCode.POWER_ON, mode)
        if res == 1:
            self.get_limit_switch()
        time.sleep(1)
        return res

    def power_off(self):
        """Robot power outage"""
        with self.lock:
            self.read_command.clear()
        return self._mesg(ProtocolCode.POWER_OFF)

    def release_all_servos(self):
        """Relax all joints
        """
        return self._mesg(ProtocolCode.RELEASE_ALL_SERVOS)

    def focus_servo(self, joint_id):
        """Power on designated servo

        Args:
            joint_id: int. joint id 1 - 7
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.FOCUS_SERVO, joint_id)

    def release_servo(self, joint_id):
        """Power off designated servo

        Args:
            joint_id: int. joint id 1 - 7
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.RELEASE_SERVO, joint_id)

    def get_robot_type(self):
        """Get robot type
        """
        return self._mesg(ProtocolCode.GET_ROBOT_ID)

    def get_zero_pos(self):
        """Read the zero encoder value

        Returns:
            list: The values of the zero encoders of the seven joints
        """
        return self._mesg(ProtocolCode.GET_ZERO_POS)

    def is_init_calibration(self):
        """Check if the robot is initialized for calibration

        Returns:
            bool: True if the robot is initialized for calibration, False otherwise
        """
        return self._mesg(ProtocolCode.IS_INIT_CALIBRATION)

    def set_break(self, joint_id, value):
        """Set break point

        Args:
            joint_id: int. joint id 1 - 7
            value: int. 0 - disable, 1 - enable

        Return:
            0 : failed
            1 : success 
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, value=value)
        return self._mesg(ProtocolCode.SET_BREAK, joint_id, value)

    def over_limit_return_zero(self):
        """Return to zero when the joint is over the limit
        """
        return self._mesg(ProtocolCode.OVER_LIMIT_RETURN_ZERO, has_reply=True)

    def jog_increment_angle(self, joint_id, increment, speed, _async=False):
        """Single angle incremental motion control. 

        Args:
            joint_id: Joint id 1 - 7.
            increment: 
            speed: int (1 - 100)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, speed=speed)
        
        return self._mesg(ProtocolCode.JOG_INCREMENT, joint_id, [self._angle2int(increment)], speed, has_reply=True, _async=_async)

    def jog_increment_coord(self, coord_id, increment, speed, _async=False):
        """Single coordinate incremental motion control. This interface is based on a single arm 1-axis coordinate system. If you are using a dual arm robot, it is recommended to use the job_base_increment_coord interface

        Args:
            joint_id: axis id 1 - 6.
            increment: 
            speed: int (1 - 100)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=coord_id, speed=speed)
        value = self._coord2int(
            increment) if coord_id <= 3 else self._angle2int(increment)
        return self._mesg(ProtocolCode.JOG_INCREMENT_COORD, coord_id, [value], speed, has_reply=True, _async=_async)
    
    def get_quick_move_info(self):
        return self._mesg(ProtocolCode.GET_QUICK_INFO)

    def drag_teach_clean(self):
        """clear sample
        """
        return self._mesg(ProtocolCode.MERCURY_DRAG_TEACH_CLEAN)

    def get_comm_error_counts(self, joint_id, _type):
        """Read the number of communication exceptions

        Args:
            joint_id (int): joint ID
            _type (int): Error type to be read, 1 ~ 4.
                1-The number of exceptions sent by the joint
                2-The number of exceptions obtained by the joint
                3-The number of exceptions sent by the end
                4-The number of exceptions read by the end
        """
        self.calibration_parameters(class_name=self.__class__.__name__, joint_id=joint_id, _type=_type)
        return self._mesg(ProtocolCode.MERCURY_ERROR_COUNTS, joint_id, _type)

    def set_pos_over_shoot(self, shoot_value):
        self.calibration_parameters(class_name=self.__class__.__name__, shoot_value=shoot_value)
        return self._mesg(ProtocolCode.MERCURY_SET_POS_OVER_SHOOT, [shoot_value*100])

    def get_pos_over_shoot(self):
        return self._mesg(ProtocolCode.MERCURY_GET_POS_OVER_SHOOT)

    def stop(self, deceleration=0, _async=False):
        """Robot stops moving

        Args:
            deceleration (bool, optional): Whether to slow down and stop. Defaults to False.

        Returns:
            int: 1 - Stop completion
        """
        self.calibration_parameters(class_name=self.__class__.__name__, deceleration=deceleration)
        # self.is_stop = time.time()
        if deceleration == 1:
            return self._mesg(ProtocolCode.STOP, 1, _async=_async)
        else:
            return self._mesg(ProtocolCode.STOP, _async=_async)

    def pause(self, deceleration=0):
        """Robot pauses movement

        Args:
            deceleration (bool, optional): Whether to slow down and stop. Defaults to False.

        Returns:
            int: 1 - pause completion
        """
        self.calibration_parameters(class_name=self.__class__.__name__, deceleration=deceleration)
        
        # self.is_stop = time.time()
        if deceleration == 1:
            return self._mesg(ProtocolCode.PAUSE, 1)
        else:
            return self._mesg(ProtocolCode.PAUSE)

    def get_modified_version(self):
        return self._mesg(ProtocolCode.ROBOT_VERSION)

    def get_pos_over(self):
        return self._mesg(ProtocolCode.GET_POS_OVER)

    def clear_encoders_error(self, joint_id):
        return self._mesg(ProtocolCode.CLEAR_ENCODERS_ERROR, joint_id)

    def get_down_encoders(self):
        return self._mesg(ProtocolCode.GET_DOWN_ENCODERS)

    def set_control_mode(self, mode):
        """Set robot motion mode

        Args:
            mode (int): 0 - location mode, 1 - torque mode

        """
        self.calibration_parameters(class_name=self.__class__.__name__, mode=mode)
        return self._mesg(ProtocolCode.SET_CONTROL_MODE, mode)

    def get_control_mode(self):
        """Get robot motion mode

        Returns:
            int: 0 - location mode, 1 - torque mode
        """
        return self._mesg(ProtocolCode.GET_CONTROL_MODE)

    def set_collision_mode(self, mode):
        """Set collision detection mode

        Args:
            mode (int): 0 - disable, 1 - enable

        """
        self.calibration_parameters(class_name=self.__class__.__name__, mode=mode)
        return self._mesg(ProtocolCode.SET_COLLISION_MODE, mode)

    def set_collision_threshold(self, joint_id, threshold_value=100):
        """Set joint collision threshold

        Args:
            joint_id (int): joint ID， range 1 ~ 6
            threshold_value (int): Collision threshold, range is 50 ~ 250, default is 100, the smaller the value, the easier it is to trigger a collision
        """
        self.calibration_parameters(class_name=self.__class__.__name__, joint_id=joint_id, threshold_value=threshold_value)
        return self._mesg(ProtocolCode.SET_COLLISION_THRESHOLD, joint_id, threshold_value)

    def get_collision_threshold(self):
        """Get joint collision threshold
        """
        return self._mesg(ProtocolCode.GET_COLLISION_THRESHOLD)

    def set_torque_comp(self, joint_id, comp_value=100):
        """Set joint torque compensation

        Args:
            joint_id (int): joint ID， range 1 ~ 6
            comp_value (int): Compensation value, range is 0 ~ 250, default is 100, The smaller the value, the harder it is to drag the joint
        """
        self.calibration_parameters(class_name=self.__class__.__name__, joint_id=joint_id, comp_value=comp_value)
        return self._mesg(ProtocolCode.SET_TORQUE_COMP, joint_id, comp_value)

    def get_torque_comp(self):
        """Get joint torque compensation
        """
        return self._mesg(ProtocolCode.GET_TORQUE_COMP)

    def power_on_only(self):
        """Only turn on the power
        """
        return self._mesg(ProtocolCode.POWER_ON_ONLY)

    def get_vr_mode(self):
        """Check if the robot is in VR mode
        """
        return self._mesg(ProtocolCode.GET_VR_MODE)

    def set_vr_mode(self, mode):
        """Set VR mode

        Args:
            mode (int): 0 - open, 1 - close
        """
        return self._mesg(ProtocolCode.SET_VR_MODE, mode)

    def get_model_direction(self):
        """Get the direction of the robot model
        """
        return self._mesg(ProtocolCode.GET_MODEL_DIRECTION)

    def set_model_direction(self, joint_id, direction):
        """Set the direction of the robot model

        Args:
            joint_id (int): joint ID, 1 ~ 7.
            direction (int): 0 - forward, 1 - backward
        """
        return self._mesg(ProtocolCode.SET_MODEL_DIRECTION, joint_id, direction)

    def get_filter_len(self, rank):
        """Get the filter length

        Args:
            rank (int): 
                1 : Drag teaching sampling filter
                2 : Drag teaching execution filter
                3 : Joint velocity fusion filter
                4 : Coordinate velocity fusion filter
                5 : Drag teaching sampling period
        """
        return self._mesg(ProtocolCode.GET_FILTER_LEN, rank)

    def set_filter_len(self, rank, value):
        """Set the filter length

        Args:
            rank (int): 
                1 : Drag teaching sampling filter
                2 : Drag teaching execution filter
                3 : Joint velocity fusion filter
                4 : Coordinate velocity fusion filter
                5 : Drag teaching sampling period
            value (int): Filter length, range is 1 ~ 100
        """
        return self._mesg(ProtocolCode.SET_FILTER_LEN, rank, value)

    def clear_zero_pos(self):
        return self._mesg(ProtocolCode.CLEAR_ZERO_POS)

    @restrict_serial_port
    def set_servo_cw(self, head_id, err_angle):
        """Set the joint in-place feedback error angle

        Args:
            head_id (int): Joint ID, 11 or 12.
            err_angle (float): Error range is 0 ~ 5.
        """
        self.calibration_parameters(class_name=self.__class__.__name__, head_id=head_id, err_angle=err_angle)
        return self._mesg(ProtocolCode.SET_SERVO_CW, head_id, [self._angle2int(err_angle)])

    @restrict_serial_port
    def get_servo_cw(self, head_id):
        """Get the joint in-place feedback error angle

        Args:
            head_id (int): Joint ID, 11 or 12.

        Returns:
            float: Error angle, range is 0 ~ 5.
        """
        self.calibration_parameters(class_name=self.__class__.__name__, head_id=head_id)
        return self._mesg(ProtocolCode.GET_SERVO_CW, head_id)

    @restrict_serial_port
    def clear_waist_queue(self):
        """Clear the cache points of the three motors in the torso
        """
        return self._mesg(ProtocolCode.CLEAR_WAIST_QUEUE)

    def clear_error_information(self):
        """Clear robot error message"""
        return self._mesg(ProtocolCode.CLEAR_ERROR_INFO)

    def get_error_information(self):
        """Obtaining robot error information

        Return:
            0: No error message.
            1 ~ 6: The corresponding joint exceeds the limit position.
            16 ~ 19: Collision protection.
            32: Kinematics inverse solution has no solution.
            33 ~ 34: Linear motion has no adjacent solution.
        """
        return self._mesg(ProtocolCode.GET_ERROR_INFO)

    def send_angles(self, angles, speed,_async=False):
        """Send the angles of all joints to robot arm.

        Args:
            angles: a list of angle values(List[float]). len 7.
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, angles=angles, speed=speed)
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.SEND_ANGLES, angles, speed, has_reply=True,_async=_async)

    def send_angle(self, joint_id, angle, speed, _async=False):
        """Send one angle of joint to robot arm.

        Args:
            joint_id : Joint id(genre.Angle)， int 1-7.
            angle : angle value(float).
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, angle=angle, speed=speed)
        return self._mesg(ProtocolCode.SEND_ANGLE, joint_id, [self._angle2int(angle)], speed, has_reply=True, _async=_async)

    def send_coord(self, coord_id, coord, speed, _async=False):
        """Send one coord to robot arm.

        Args:
            coord_id (int): coord id, range 1 ~ 6
            coord (float): coord value.
                The coord range of `X` is -351.11 ~ 566.92.
                The coord range of `Y` is -645.91 ~ 272.12.
                The coord range of `Y` is -262.91 ~ 655.13.
                The coord range of `RX` is -180 ~ 180.
                The coord range of `RY` is -180 ~ 180.
                The coord range of `RZ` is -180 ~ 180.
            speed (int): 1 ~ 100
        """

        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=coord_id, coord=coord, speed=speed)
        value = self._coord2int(coord) if coord_id <= 3 else self._angle2int(coord)
        return self._mesg(ProtocolCode.SEND_COORD, coord_id, [value], speed, has_reply=True, _async=_async)

    def send_coords(self, coords, speed, _async=False):
        """Send all coords to robot arm.

        Args:
            coords: a list of coords value(List[float]). len 6 [x, y, z, rx, ry, rz]
                The coord range of `X` is -351.11 ~ 566.92.
                The coord range of `Y` is -645.91 ~ 272.12.
                The coord range of `Y` is -262.91 ~ 655.13.
                The coord range of `RX` is -180 ~ 180.
                The coord range of `RY` is -180 ~ 180.
                The coord range of `RZ` is -180 ~ 180.
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coords=coords, speed=speed)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SEND_COORDS, coord_list, speed, has_reply=True, _async=_async)

    def resume(self):
        """Recovery movement"""
        return self._mesg(ProtocolCode.RESUME)

    def set_servo_calibration(self, joint_id):
        """The current position of the calibration joint actuator is the angle zero point, 

        Args:
            joint_id: Serial number of articulated steering gear. Joint id 1 - 6
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.SET_SERVO_CALIBRATION, joint_id)
    
    def set_servos_calibration(self):
        for id in range(1, 8):
            self._mesg(ProtocolCode.SET_SERVO_CALIBRATION, id)

    def is_in_position(self, data, mode=0):
        """Judge whether in the position.

        Args:
            data: A data list, angles or coords. angles len 6, coords len 6.
            mode: 1 - coords, 0 - angles

        Return:
            1 - True\n
            0 - False\n
            -1 - Error
        """
        if mode == 1:
            self.calibration_parameters(
                class_name=self.__class__.__name__, coords=data)
            data_list = []
            for idx in range(3):
                data_list.append(self._coord2int(data[idx]))
            for idx in range(3, 6):
                data_list.append(self._angle2int(data[idx]))
        elif mode == 0:
            self.calibration_parameters(
                class_name=self.__class__.__name__, angles=data)
            data_list = [self._angle2int(i) for i in data]
        else:
            raise Exception("mode is not right, please input 0 or 1")

        return self._mesg(ProtocolCode.IS_IN_POSITION, data_list, mode)

    def is_moving(self):
        """Detect if the robot is moving

        Return:
            0 - not moving
            1 - is moving
            -1 - error data
        """
        return self._mesg(ProtocolCode.IS_MOVING)

    def jog_angle(self, joint_id, direction, speed, _async=True):
        """Jog control angle.

        Args:
            joint_id (int): Joint id 1 - 7.
            direction (int): 0 - decrease, 1 - increase
            speed (int): int range 1 - 100
            _async (bool, optional): Whether to execute asynchronous control. Defaults to True.

        Return:
            0: End of exercise.
            1 ~ 7: Corresponding joint exceeds limit.
            16 ~ 19: Collision protection.
            32 ~ 35: Coordinates have no adjacent solutions.
            49: Not powered on.
            50: Motor abnormality.
            51: Motor encoder error
            52: Not reaching the designated location or not reaching the designated location for more than 5 minutes (only J11, J12 available)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, direction=direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_ANGLE, joint_id, direction, speed, _async=_async, has_reply=True)

    def jog_coord(self, coord_id, direction, speed, _async=True):
        """Jog control coord. This interface is based on a single arm 1-axis coordinate system. If you are using a dual arm robot, it is recommended to use the jog_base_coord interface

        Args:
            coord_id (int): int 1-6
            direction (int): 0 - decrease, 1 - increase
            speed (int): 1 - 100
            _async (bool, optional): Whether to execute asynchronous control. Defaults to True.

        Returns:
            1: End of the Movement

        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=coord_id, direction=direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_COORD, coord_id, direction, speed, _async=_async, has_reply=True)

    @restrict_serial_port
    def jog_base_increment_coord(self, axis_id, increment, speed, _async=True):
        """Single coordinate incremental motion control

        Args:
            axis_id (int): axis id, range 1 ~ 6 corresponds to [x,y,z,rx,ry,rz]
            increment (float): Incremental value
            speed (int): speed
            
        Return:
            0: End of exercise.
            1 ~ 7: Corresponding joint exceeds limit.
            16 ~ 19: Collision protection.
            32 ~ 35: Coordinates have no adjacent solutions.
            49: Not powered on.
            50: Motor abnormality.
            51: Motor encoder error
            52: Not reaching the designated location or not reaching the designated location for more than 5 minutes (only J11, J12 available)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=axis_id, speed=speed)
        coord_list = []
        if axis_id < 4:
            coord_list.append(self._coord2int(increment))
        else:
            coord_list.append(self._angle2int(increment))
        return self._mesg(ProtocolCode.JOG_BASE_INCREMENT_COORD, axis_id, coord_list, speed, has_reply=True, _async=_async)

    def get_max_speed(self, mode):
        """Get maximum speed
        
        Args:
            mode (int): 0 - angle speed. 1 - coord speed.
            
        Return: 
            angle speed range 1 ~ 150°/s. coord speed range 1 ~ 200mm/s
        """
        self.calibration_parameters(class_name=self.__class__.__name__, mode=mode)
        return self._mesg(ProtocolCode.GET_SPEED, mode)

    def set_max_speed(self, mode, max_speed):
        """Set maximum speed

        Args:
            mode (int): 0 - angle speed. 1 - coord speed.
            max_speed (int): angle speed range 1 ~ 150°/s. coord speed range 1 ~ 200mm/s

        Returns:
            1: _description_
        """
        self.calibration_parameters(lass_name=self.__class__.__name__, mode=mode, max_speed=max_speed)
        return self._mesg(ProtocolCode.SET_SPEED, mode, [max_speed])

    def set_max_acc(self, mode, max_acc):
        """Set maximum acceleration

        Args:
            mode (int): 0 - angle acceleration. 1 - coord acceleration.
            max_acc (int): maximum acceleration value. Angular acceleration range is 1 ~ 150°/s. Coordinate acceleration range is 1 ~ 400mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, mode=mode, max_acc=max_acc)
        return self._mesg(ProtocolCode.SET_MAX_ACC, mode, [max_acc])

    def get_max_acc(self, mode):
        """Get maximum acceleration

        Args:
            mode (int): 0 - angle acceleration. 1 - coord acceleration.
        """
        return self._mesg(ProtocolCode.GET_MAX_ACC, mode)

    def get_joint_min_angle(self, joint_id):
        """Gets the minimum movement angle of the specified joint

        Args: 
            joint_id: Joint id 1 - 6 or 11 ~ 12

        Return:
            angle value(float)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MIN_ANGLE, joint_id)

    def get_joint_max_angle(self, joint_id):
        """Gets the maximum movement angle of the specified joint

        Args:
            joint_id: Joint id 1 - 6 or 11 ~ 12

        Return:
            angle value(float)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MAX_ANGLE, joint_id)

    def set_joint_max_angle(self, joint_id, degree):
        """Set the maximum angle of the joint (must not exceed the maximum angle specified for the joint)

        Args:
            joint_id (int): Joint id 1 - 6  or 11 ~ 12
            degree: The angle range of joint 1 is -165 ~ 165. The angle range of joint 2 is -55 ~ 95. The angle range of joint 3 is -173 ~ 5. The angle range of joint 4 is -165 ~ 165. The angle range of joint 5 is -20 ~ 265. The angle range of joint 6 is -180 ~ 180. The angle range of joint 11 is -60 ~ 0. The angle range of joint 12 is -138 ~ 188.

        Return:
            1 - success
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, degree=degree)
        return self._mesg(ProtocolCode.SET_JOINT_MAX, joint_id, degree)

    def set_joint_min_angle(self, joint_id, degree):
        """Set the minimum angle of the joint (must not be less than the minimum angle specified by the joint)

        Args:
            joint_id (int): Joint id 1 - 6.
            degree: The angle range of joint 1 is -165 ~ 165. The angle range of joint 2 is -55 ~ 95. The angle range of joint 3 is -173 ~ 5. The angle range of joint 4 is -165 ~ 165. The angle range of joint 5 is -20 ~ 265. The angle range of joint 6 is -180 ~ 180. The angle range of joint 11 is -60 ~ 0. The angle range of joint 12 is -138 ~ 188.

        Return:
            1 - success
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, degree=degree)
        return self._mesg(ProtocolCode.SET_JOINT_MIN, joint_id, degree)

    def is_servo_enable(self, joint_id):
        """To detect the connection state of a single joint

        Args:
            joint_id: Joint id 1 - 7

        Return:
            0 - disable
            1 - enable
            -1 - error
        """
        self.calibration_parameters(class_name = self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.IS_SERVO_ENABLE, joint_id)

    def is_all_servo_enable(self):
        """Detect the connection status of all joints

        Return:
            0 - disable
            1 - enable
            -1 - error
        """
        return self._mesg(ProtocolCode.IS_ALL_SERVO_ENABLE)

    def get_servo_data(self, servo_id, address, mode):
        """Read the data parameter of the specified address of the steering gear.

        Args:
            servo_id: Joint id 11 or 12
            address: Data address. range 5 ~ 69
            mode: 1 - indicates that value is one byte(default), 2 - represents a value of two bytes.

        Return:
            values 0 - 4096
        """
        if mode not in [1, 2]:
            raise Exception("mode must be 1 or 2")
        self.calibration_parameters(class_name = self.__class__.__name__, servo_id=servo_id)
        return self._mesg(
            ProtocolCode.GET_SERVO_DATA, servo_id, address, mode
        )
        
    def set_servo_data(self, servo_id, address, value, mode):
        """Set the data parameters of the specified address of the steering gear

        Args:
            servo_id: Joint id 11 or 12
            address: Data address. range 5 ~ 69
            value: 0 - 4096
            mode: 1 - indicates that value is one byte(default), 2 - represents a value of two bytes.
        """
        if mode not in [1, 2]:
            raise Exception("mode must be 1 or 2")
        self.calibration_parameters(class_name = self.__class__.__name__, servo_id=servo_id)
        if mode == 1:
            return self._mesg(ProtocolCode.SET_SERVO_DATA, servo_id, address, value, mode)
        else:
            return self._mesg(ProtocolCode.SET_SERVO_DATA, servo_id, address, [value], mode)

    def get_servo_speeds(self):
        """Get joint speed
        
        Return: 
            unit step/s
        """
        return self._mesg(ProtocolCode.GET_SERVO_SPEED)
    
    def get_servo_currents(self):
        """Get joint current
        
        Return: 
            0 ~ 5000 mA
        """
        return self._mesg(ProtocolCode.GET_SERVO_CURRENTS)
    
    def get_servo_status(self):
        """Get joint status
        
        """
        return self._mesg(ProtocolCode.GET_SERVO_STATUS)
    
    def set_gripper_state(self, flag, speed, gripper_type=None):
        """Set gripper switch state

        Args:
            flag  (int): 0 - open, 1 - close
            speed (int): 1 ~ 100
            gripper_type (int): default 1
                1 : Adaptive gripper. default to adaptive gripper
                2 : Parallel Gripper
        """
        if gripper_type is None:
            self.calibration_parameters(class_name = self.__class__.__name__, flag=flag, speed=speed)
            return self._mesg(ProtocolCode.SET_GRIPPER_STATE, flag, speed)
        else:
            self.calibration_parameters(class_name = self.__class__.__name__, flag=flag, speed=speed, gripper_type=gripper_type)
            return self._mesg(ProtocolCode.SET_GRIPPER_STATE, flag, speed, gripper_type)
        
    def set_gripper_value(self, gripper_value, speed, gripper_type=None):
        """Set gripper value

        Args:
            gripper_value (int): 0 ~ 100
            speed (int): 1 ~ 100
            gripper_type (int): default 1
                1 : Adaptive gripper. default to adaptive gripper
                2 : Parallel Gripper
        """
        
        if gripper_type is not None:
            self.calibration_parameters(class_name=self.__class__.__name__, gripper_value=gripper_value, speed=speed,
                                        gripper_type=gripper_type)
            return self._mesg(ProtocolCode.SET_GRIPPER_VALUE, gripper_value, speed, gripper_type)
        else:
            self.calibration_parameters(class_name=self.__class__.__name__, gripper_value=gripper_value, speed=speed)
            return self._mesg(ProtocolCode.SET_GRIPPER_VALUE, gripper_value, speed)
        
    def set_gripper_calibration(self):
        """Set the current position to zero, set current position value is `2048`."""
        return self._mesg(ProtocolCode.SET_GRIPPER_CALIBRATION)
    
    def set_gripper_mode(self, mode):
        """Set gripper mode
        
        Args:
            mode: 0 - transparent transmission. 1 - Port Mode.
        
        """
        self.calibration_parameters(class_name = self.__class__.__name__, mode=mode)
        return self._mesg(ProtocolCode.SET_GRIPPER_MODE, mode)
    
    def get_gripper_mode(self):
        """Get gripper mode
        
        Return:
            mode: 0 - transparent transmission. 1 - Port Mode.
        """
        return self._mesg(ProtocolCode.GET_GRIPPER_MODE)
    
    def set_color(self, r=0, g=0, b=0):
        """Set the light color on the top of the robot end.

        Args:
            r (int): 0 ~ 255
            g (int): 0 ~ 255
            b (int): 0 ~ 255

        """
        self.calibration_parameters(class_name = self.__class__.__name__, rgb=[r, g, b])
        return self._mesg(ProtocolCode.SET_COLOR, r, g, b)
    
    def set_digital_output(self, pin_no, pin_signal):
        """Set the end-of-arm IO status

        Args:
            pin_no (int): 1 or 2
            pin_signal (int): 0 / 1
        """
        self.calibration_parameters(class_name = self.__class__.__name__, pin_no=pin_no, pin_signal=pin_signal)
        return self._mesg(ProtocolCode.SET_DIGITAL_OUTPUT, pin_no, pin_signal)

    def get_digital_input(self, pin_no):
        """Read the end-of-arm IO status

        Args:
            pin_no (int): 1 or 2

        Returns:
            int: 0 or 1
        """
        self.calibration_parameters(class_name = self.__class__.__name__, pin_no=pin_no)
        return self._mesg(ProtocolCode.GET_DIGITAL_INPUT, pin_no)

    def get_digital_inputs(self):
        """Read the end-of-arm IO status, IN1/IN2/ATOM Button"""
        return self._mesg(ProtocolCode.GET_DIGITAL_INPUTS)

    def get_world_reference(self):
        """Get the world coordinate system"""
        return self._mesg(ProtocolCode.GET_WORLD_REFERENCE)
    
    def set_world_reference(self, coords):
        """Set the world coordinate system
        
        Args:
            coords: a list of coords value(List[float]). [x(mm), y, z, rx(angle), ry, rz]
        """
        # self.calibration_parameters(class_name = self.__class__.__name__, coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_WORLD_REFERENCE, coord_list)
    
    def set_tool_reference(self, coords):
        """Set tool coordinate system
        
        Args:
            coords: a list of coords value(List[float])
        """
        # self.calibration_parameters(class_name = self.__class__.__name__, coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_TOOL_REFERENCE, coord_list)
    
    def get_tool_reference(self):
        """Get tool coordinate system """
        return self._mesg(ProtocolCode.GET_TOOL_REFERENCE)
    
    def set_reference_frame(self, rftype):
        """Set the base coordinate system
        
        Args:
            rftype: 0 - base 1 - tool.
        """
        self.calibration_parameters(class_name = self.__class__.__name__, rftype=rftype)
        return self._mesg(ProtocolCode.SET_REFERENCE_FRAME, rftype)
    
    def get_reference_frame(self):
        """Get the base coordinate system
        
        Return: 
            0 - base 1 - tool.
        """
        return self._mesg(ProtocolCode.GET_REFERENCE_FRAME)
    
    def set_movement_type(self, move_type):
        """Set movement type
        
        Args:
            move_type: 2 - movel, 0 - moveJ
        """
        self.calibration_parameters(class_name = self.__class__.__name__, move_type=move_type)
        return self._mesg(ProtocolCode.SET_MOVEMENT_TYPE, move_type)
    
    def get_movement_type(self):
        """Get movement type
        
        Return: 
            1 - movel, 0 - moveJ
        """
        return self._mesg(ProtocolCode.GET_MOVEMENT_TYPE)
    
    def set_end_type(self, end):
        """Set end coordinate system
        
        Args:
            end: int
                0 - flange, 1 - tool
        """
        self.calibration_parameters(class_name = self.__class__.__name__, end=end)
        return self._mesg(ProtocolCode.SET_END_TYPE, end)
    
    def get_end_type(self):
        """Get end coordinate system
        
        Return: 
            0 - flange, 1 - tool
        """
        return self._mesg(ProtocolCode.GET_END_TYPE)
        
    def get_monitor_mode(self):
        return self._mesg(ProtocolCode.GET_MONITOR_MODE)
    
    def set_monitor_mode(self,mode):
        return self._mesg(ProtocolCode.SET_MONITOR_MODE, mode)
    
    def set_limit_switch(self, limit_mode, state):
        """Set the limit switches. No save after power off

        Args:
            limit_mode (int): 1 - Location out of tolerance. 2 - Synchronous control
            state (int): 0 - close. 1 - open
        """
        self.calibration_parameters(class_name = self.__class__.__name__, limit_mode=limit_mode, state=state)
        if limit_mode == 2 and state == 0:
            self.sync_mode = False
        elif limit_mode == 2 and state == 1:
            self.sync_mode = True
        return self._mesg(ProtocolCode.SET_LIMIT_SWITCH, limit_mode, state)
    
    def get_limit_switch(self):
        """Get the limit switches
        
        Return:
            list : [Location out of tolerance state, sync state], 0 - close. 1 - open
        """
        res = self._mesg(ProtocolCode.GET_LIMIT_SWITCH)
        if isinstance(res, list) and res[1] == 0:
            self.sync_mode = False
        return res
    
    def solve_inv_kinematics(self, new_coords, old_angles):
        """_summary_
        """
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(new_coords[idx]))
        for angle in new_coords[3:]:
            coord_list.append(self._angle2int(angle))
        angles = [self._angle2int(angle) for angle in old_angles]
        return self._mesg(ProtocolCode.SOLVE_INV_KINEMATICS, coord_list, angles, has_reply=True)
    
    def get_drag_fifo(self):
        return self._mesg(ProtocolCode.GET_DRAG_FIFO)
    
    def set_drag_fifo(self, angles):
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.SET_DRAG_FIFO, angles)
    
    def get_drag_fifo_len(self):
        return self._mesg(ProtocolCode.GET_DRAG_FIFO_LEN)
    
    def jog_rpy(self, axis, direction, speed):
        """Rotate the end point around the fixed axis of the base coordinate system

        Args:
            axis (int): 1 ~ 3. 1 - Roll, 2 - Pitch, 3 - Yaw
            direction (int): 1 - Forward. 0 - Reverse.
            speed (int): 1 ~ 100.
        """
        self.calibration_parameters(class_name = self.__class__.__name__, axis=axis, direction=direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_RPY, axis, direction, speed, has_reply=True)
    
    def get_collision_mode(self):
        """Get collision detection status
        
        Return:
            1 - open
            0 - close
        """
        return self._mesg(ProtocolCode.GET_COLLISION_MODE)
    
    def set_electric_gripper(self, mode):
        """Set the state of the electric gripper

        Args:
            mode (int): 0 - open. 1 - close
        """
        self.calibration_parameters(class_name = self.__class__.__name__, mode=mode)
        return self._mesg(ProtocolCode.SET_ELECTRIC_GRIPPER, mode)
    
    def init_electric_gripper(self):
        """Electric Gripper Initialization
        """
        return self._mesg(ProtocolCode.INIT_ELECTRIC_GRIPPER)
    
    def get_servo_encoder(self, id):
        return self._mesg(ProtocolCode.GET_ENCODER, id)
    
    def get_servo_encoders(self):
        return self._mesg(ProtocolCode.GET_ENCODERS)
    
    def set_base_io_output(self, pin_no, pin_signal):
        """Set the base output IO status

        Args:
            pin_no: pin port number. range 1 ~ 6
            pin_signal: 0 - low. 1 - high.
        """
        self.calibration_parameters(class_name = self.__class__.__name__, pin_no=pin_no, pin_signal=pin_signal)
        return self._mesg(ProtocolCode.SET_BASIC_OUTPUT, pin_no, pin_signal)

    def get_base_io_input(self, pin_no):
        """Get the input IO status of the base

        Args:
            pin_no: (int) pin port number. range 1 ~ 6
        """
        self.calibration_parameters(class_name = self.__class__.__name__, pin_no=pin_no)
        return self._mesg(ProtocolCode.GET_BASIC_INPUT, pin_no)

    def set_world_reference(self, coords):
        """Set the world coordinate system
        
        Args:
            coords: a list of coords value(List[float])
                for mycobot / mecharm / myArm: [x(mm), y, z, rx(angle), ry, rz]\n
        """
        self.calibration_parameters(class_name = self.__class__.__name__, coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_WORLD_REFERENCE, coord_list)
    
    def set_identify_mode(self, mode):
        return self._mesg(ProtocolCode.SET_IDENTIFY_MODE, mode)
    
    def get_identify_mode(self):
        return self._mesg(ProtocolCode.GET_IDENTIFY_MODE)
    
    def write_move_c_r(self, coords, r, speed, rank=0):
        """_summary_

        Args:
            coords (_type_): _description_
            r (_type_): _description_
            speed (_type_): _description_
            rank (_type_): _description_
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coords=coords, r=r, speed=speed, rank=rank)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.WRITE_MOVE_C_R, coord_list, [r*100], speed, rank, has_reply=True)

    def fourier_trajectories(self, trajectory):
        self.calibration_parameters(class_name=self.__class__.__name__, trajectory=trajectory)
        return self._mesg(ProtocolCode.FOURIER_TRAJECTORIES, trajectory)
    
    def get_dynamic_parameters(self, add):
        return self._mesg(ProtocolCode.GET_DYNAMIC_PARAMETERS, add)
    
    def set_dynamic_parameters(self, add, data):
        return self._mesg(ProtocolCode.SET_DYNAMIC_PARAMETERS, add, [data*1000])
    
    def identify_print(self):
        res = self.all_debug_data
        self.all_debug_data = []
        return res
    
    def clear_encoder_error(self, joint_id):
        return self._mesg(ProtocolCode.CLEAR_ENCODER_ERROR, joint_id)
    
    def set_pro_gripper(self, gripper_id, gripper_address, value=0, has_return=False):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id, gripper_address=gripper_address)
        return self._mesg(ProtocolCode.MERCURY_SET_TOQUE_GRIPPER, gripper_id, [gripper_address], [value], has_return=has_return)
    
    def get_pro_gripper(self, gripper_id, gripper_address):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id, gripper_address=gripper_address)
        return self._mesg(ProtocolCode.MERCURY_GET_TOQUE_GRIPPER, gripper_id, [gripper_address])
    
    def set_pro_gripper_angle(self, gripper_id, gripper_angle):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id, gripper_angle=gripper_angle)
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ANGLE, gripper_angle)
    
    def set_pro_gripper_open(self, gripper_id):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id)
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ANGLE, 100)
    
    def set_pro_gripper_close(self, gripper_id):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id)
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ANGLE, 0)
    
    def get_pro_gripper_angle(self, gripper_id):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id)
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_ANGLE)
    
    def set_pro_gripper_calibration(self, gripper_id):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id)
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_CALIBRATION)
    
    def get_pro_gripper_status(self, gripper_id):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id)
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_STATUS)
    
    def set_pro_gripper_torque(self, gripper_id, torque):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id, torque=torque)
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_TORQUE, torque)
    
    def get_pro_gripper_torque(self, gripper_id):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id)
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_TORQUE)
    
    def set_pro_gripper_speed(self, gripper_id, speed):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id, speed=speed)
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_SPEED, speed)
    
    def get_pro_gripper_speed(self, gripper_id):
        self.calibration_parameters(class_name = self.__class__.__name__, gripper_id=gripper_id)
        return self.get_pro_gripper(gripper_id, ProGripper.GET_GRIPPER_SPEED)
    
    def set_pro_gripper_abs_angle(self, gripper_id, angle):
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_ABS_ANGLE, angle, has_return=True)
    
    def set_pro_gripper_pause(self, gripper_id):
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_PAUSE)
    
    def set_pro_gripper_stop(self, gripper_id):
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_STOP)
    
    def set_pro_gripper_resume(self, gripper_id):
        return self.set_pro_gripper(gripper_id, ProGripper.SET_GRIPPER_RESUME)
    
    def get_motors_run_err(self):
        return self._mesg(ProtocolCode.GET_MOTORS_RUN_ERR)



#!/usr/bin/env python3
import socket
import tkinter
from tkinter import ttk
import time
import threading
import os
import textwrap
import serial
import serial.tools.list_ports

from pymycobot import MyPalletizer


LOG_NUM = 0


class MycobotTest(object):
    def __init__(self):
        self.mycobot = None

        self.win = tkinter.Tk()
        self.win.title("树莓派版 MyPalletizer 测试工具")
        self.win.geometry("918x611+10+10")  # 290 160为窗口大小，+10 +10 定义窗口弹出时的默认展示位置

        self.port_label = tkinter.Label(self.win, text="选择串口：")
        self.port_label.grid(row=0)
        self.port_list = ttk.Combobox(
            self.win, width=15, postcommand=self.get_serial_port_list
        )  # #创建下拉菜单
        self.get_serial_port_list()  # #给下拉菜单设定值
        self.port_list.current(0)
        self.stop_test = False
        self.port_list.grid(row=0, column=1)

        # self.baud_label = tkinter.Label(self.win, text="选择波特率：")
        # self.baud_label.grid(row=1)
        # self.baud_list = ttk.Combobox(self.win, width=15)
        # self.baud_list["value"] = ("115200", "115200")
        # self.baud_list.current(0)
        # self.baud_list.grid(row=1, column=1)

        # Connect
        self.connect_label = tkinter.Label(self.win, text="连接机器：")
        self.connect_label.grid(row=2)
        self.connect = tkinter.Button(self.win, text="连接", command=self.connect_mycobot)
        self.disconnect = tkinter.Button(
            self.win, text="断开", command=self.disconnect_mycobot
        )
        self.connect.grid(row=3)
        self.disconnect.grid(row=3, column=1)

        # Check servo.
        self.check_label = tkinter.Label(self.win, text="检测连接：")
        self.check_label.grid(row=4)
        self.check_btn = tkinter.Button(
            self.win, text="开始检测", command=self.check_mycobot_servos
        )
        self.check_btn.grid(row=4, column=1)

        # Calibration.
        self.calibration_num = None
        self.calibration_label = tkinter.Label(self.win, text="校准舵机：")
        self.calibration_label.grid(row=5)
        self.calibration_btn = tkinter.Button(
            self.win, text="开始校准", command=self.calibration_mycobot
        )
        self.calibration_btn.grid(row=5, column=1)

        # LED.
        self.set_color_label = tkinter.Label(self.win, text="测试Atom灯板：")
        self.set_color_label.grid(row=6, columnspan=2)
        self.color_red = tkinter.Button(
            self.win, text="设置红色", command=lambda: self.send_color("red")
        )
        self.color_green = tkinter.Button(
            self.win, text="设置绿色", command=lambda: self.send_color("green")
        )
        self.color_red.grid(row=7)
        self.color_green.grid(row=7, column=1)

        # Aging test.
        self.aging_stop = False
        self.movement_label = tkinter.Label(self.win, text="老化循环动作：")
        self.movement_label.grid(row=8)
        self.start_btn = tkinter.Button(
            self.win, text="开始", command=self.start_aging_test
        )
        self.start_btn.grid(row=9)
        self.stop_btn = tkinter.Button(
            self.win, text="停止", command=self.stop_aging_test
        )
        self.stop_btn.grid(row=9, column=1)

        # Release
        self.release_btn = tkinter.Button(
            self.win, text="放松所有电机", command=self.release_mycobot
        )
        self.release_btn.grid(row=10)

        # rectify
        self.rectify_btn = tkinter.Button(
            self.win, text="使能所有电机", command=self.rectify_mycobot
        )
        self.rectify_btn.grid(row=10, column=1)
        
        self.set_data= tkinter.Button(
            self.win, text="校准电机参数", command=self.set_servo_data
        )
        self.set_data.grid(row=11, column=0)

        # Log output.
        self.log_label = tkinter.Label(self.win, text="日志：")
        self.log_label.grid(row=0, column=12)
        _f = tkinter.Frame(self.win)
        _bar = tkinter.Scrollbar(_f, orient=tkinter.VERTICAL)
        self.log_data_Text = tkinter.Text(
            _f, width=100, height=35, yscrollcommand=_bar.set
        )
        _bar.pack(side=tkinter.RIGHT, fill=tkinter.Y)
        _bar.config(command=self.log_data_Text.yview)
        self.log_data_Text.pack()
        # self.log_data_Text.grid(row=1, column=12, rowspan=15, columnspan=10)
        _f.grid(row=1, column=12, rowspan=15, columnspan=10)

    def run(self):
        self.win.mainloop()  # run

    # ============================================================
    # Connect method
    # ============================================================
    def connect_mycobot(self):
        self.prot = port = self.port_list.get()
        if not port:
            self.write_log_to_Text("请选择串口")
            return
        # self.baud = baud = self.baud_list.get()
        # if not baud:
        #     self.write_log_to_Text("请选择波特率")
        #     return
        baud = 115200

        try:
            # self.mycobot = MyCobot(PI_PORT, PI_BAUD)
            self.mycobot = MyPalletizer(port, baud)
            # self.mycobot = MyCobot("/dev/cu.usbserial-0213245D", 115200)
            self.write_log_to_Text("连接成功 !")
        except Exception as e:
            err_log = """\
                \r连接失败 !!!
                \r=================================================
                {}
                \r=================================================
            """.format(
                e
            )
            self.write_log_to_Text(err_log)

    def disconnect_mycobot(self):
        if not self.has_mycobot():
            return

        try:
            del self.mycobot
            self.mycobot = None
            self.write_log_to_Text("断开连接成功 !")
        except AttributeError:
            self.write_log_to_Text("还没有连接myPalletizer！！！")

    # ============================================================
    #  Function method
    # ============================================================
    def release_mycobot(self):
        if not self.has_mycobot():
            return
        self.mycobot.release_all_servos()
        self.write_log_to_Text("Release over.")

    def check_mycobot_servos(self):
        if not self.has_mycobot():
            return

        # ping_commands = [
        #     [255, 255, 1, 2, 1, 251],
        #     [255, 255, 2, 2, 1, 250],
        #     [255, 255, 3, 2, 1, 249],
        #     [255, 255, 4, 2, 1, 248],
        #     [255, 255, 5, 2, 1, 247],
        #     [255, 255, 6, 2, 1, 246],
        #     [255, 255, 7, 2, 1, 246],
        # ]
        res = []
        # for idx, command in enumerate(ping_commands, start=1):
        #     self.mycobot._write(command)
        #     time.sleep(0.1)
        #     if not self.mycobot._read(1):
        #         res.append(idx)
        #     time.sleep(0.1)
        for i in range(1,5):
            r = self.mycobot.is_servo_enable(i)
            if r != 1:
                res.append(i)
            time.sleep(0.1)
        if res:
            self.write_log_to_Text("关节 {} 无法通信！！！".format(res))
        else:
            self.write_log_to_Text("所有关节连接正常。")

    def calibration_mycobot(self):
        """Calibration button click event.

        Click to calibrate one motor at a time and calibrate in turn. After all
        calibration, resume initialization.
        """
        if not self.has_mycobot():
            return

        if not self.calibration_num:
            self.calibration_num = 0

        self.calibration_num += 1

        self.mycobot.set_servo_calibration(self.calibration_num)
        time.sleep(0.1)
        self.mycobot.send_angle(self.calibration_num, 0, 20)
        time.sleep(0.1)
        self.write_log_to_Text("校准电机 %s 结束." % self.calibration_num)

        if self.calibration_num == 4:
            self.write_log_to_Text("全部校准完成.")
            self.calibration_num = None
            # self.rectify_mycobot()
            self._calibration_test()

    def send_color(self, color: str):
        if not self.has_mycobot():
            return

        color_dict = {
            "red": [255, 0, 0],
            "green": [0, 255, 0],
            "blue": [0, 0, 255],
        }
        self.mycobot.set_color(*color_dict[color])
        self.write_log_to_Text("发送颜色: {}.".format(color))
        
    def set_servo_data(self):
        if not self.has_mycobot():
            return
        address = [21, 22, 23, 24, 26, 27]
        joint_1 = [25,25,0,0,3,3]
        joint_other = [32,8,0,0,1,1]
        for i in range(1,5):
            if i == 1:
                for j in range(len(address)):
                    self.mycobot.set_servo_data(i, address[j], joint_1[j])
                    time.sleep(0.1)
            else:
                for j in range(len(address)):
                    self.mycobot.set_servo_data(i, address[j], joint_other[j])
                    time.sleep(0.1)
        for i in range(1, 5):
            for j in range(len(address)):
                res = "设置电机{}参数 {} ".format(i, address[j])
                if i == 1:
                    data = self.mycobot.get_servo_data(i, address[j])
                    if data == joint_1[j]:
                        res += "成功"
                    else:
                        res += "失败{}".format(data)
                else:
                    data = self.mycobot.get_servo_data(i, address[j])
                    if data == joint_other[j]:
                        res += "成功"
                    else:
                        res += "失败{}".format(data)
                time.sleep(0.1)
                self.write_log_to_Text(res)
        

    def start_aging_test(self):
        if not self.has_mycobot():
            return

        self.aging_stop = False
        self.aging = threading.Thread(target=self._aging_test, daemon=True)
        self.aging.start()
        # self._aging_test()
        self.write_log_to_Text("开始循环老化测试 ...")

    def stop_aging_test(self):
        try:
            # os.system("sudo systemctl stop aging_test.service")
            # os.system("sudo rm /home/er/aging_test.sh")
            # os.system("sudo rm /home/er/Desktop/aging_test.py")
            # os.system("sudo rm /etc/systemd/system/aging_test.service")
            # os.system("sudo systemctl daemon-reload")
            self.stop_test = True
            self.write_log_to_Text("结束循环老化测试.")
        except:
            self.write_log_to_Text("结束老化测试失败 ！！！")

    def rectify_mycobot(self):
        if not self.has_mycobot():
            return
        for i in range(1,5):
            self.mycobot.focus_servo(i)
            time.sleep(0.1)
        self.write_log_to_Text("使能所有电机完成.")

    # ============================================================
    # Utils method
    # ============================================================
    def has_mycobot(self):
        """Check whether it is connected on mycobot"""
        if not self.mycobot:
            self.write_log_to_Text("还没有连接机器！！！")
            return False
        return True

    def _aging_test(self):
        """
        Aging test thread target.
        By using in `start_aging_test()` and `stop_aging_test()`.
        """
        # if socket.gethostname() != "pi":
        #     self.write_log_to_Text("老化测试支持 Raspberry OS.")
        #     return
        start = time.time()
        self.stop_test = False
        while True and time.time() - start < 60*30 and self.stop_test == False:
            self.mycobot.set_color(255, 0, 0)
            self.mycobot.wait(5).send_angles([0, 0, 0, 0], 95)
            self.mycobot.wait(3).send_angles([-160, 0, 0, 0], 95)
            self.mycobot.wait(3).send_angles([160, 0, 0, 0], 95)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 95)
            self.mycobot.wait(3).send_angles([0, 80, 0, 0], 95)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 95)
            self.mycobot.wait(3).send_angles([0, 0, 20, 0], 95)
            self.mycobot.wait(3).send_angles([0, 0, 40, 0], 95)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 95)
            self.mycobot.wait(3).send_angles([0, 0, 0, 180], 95)
            self.mycobot.wait(3).send_angles([0, 0, 0, -180], 95)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 95)

            # middle
            self.mycobot.set_color(0, 255, 0)
            self.mycobot.wait(5).send_angles([0, 0, 0, 0], 55)
            self.mycobot.wait(5).send_angles([-160, 0, 0, 0], 55)
            self.mycobot.wait(5).send_angles([160, 0, 0, 0], 55)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 55)
            self.mycobot.wait(3).send_angles([0, 80, 0, 0], 55)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 55)
            self.mycobot.wait(3).send_angles([0, 0, 20, 0], 55)
            self.mycobot.wait(3).send_angles([0, 0, 40, 0], 55)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 55)
            self.mycobot.wait(6).send_angles([0, 0, 0, 180], 55)
            self.mycobot.wait(6).send_angles([0, 0, 0, -180], 55)
            self.mycobot.wait(6).send_angles([0, 0, 0, 0], 55)

            # slow
            self.mycobot.set_color(0, 0, 255)
            self.mycobot.wait(5).send_angles([0, 0, 0, 0], 15)
            self.mycobot.wait(8).send_angles([-160, 0, 0, 0], 15)
            self.mycobot.wait(8).send_angles([160, 0, 0, 0], 15)
            self.mycobot.wait(5).send_angles([0, 0, 0, 0], 15)
            self.mycobot.wait(3).send_angles([0, 80, 0, 0], 15)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 15)
            self.mycobot.wait(3).send_angles([0, 0, 20, 0], 15)
            self.mycobot.wait(3).send_angles([0, 0, 40, 0], 15)
            self.mycobot.wait(3).send_angles([0, 0, 0, 0], 15)
            self.mycobot.wait(8).send_angles([0, 0, 0, 180], 15)
            self.mycobot.wait(8).send_angles([0, 0, 0, -180], 15)
            self.mycobot.wait(5).send_angles([0, 0, 0, 0], 15)
        self.mycobot.send_angles([0, 0, 0, 0], 35)
        self.write_log_to_Text("老化测试结束。")
        # aging_test_content_py = textwrap.dedent(
        #     """\
        #     #!/usr/bin/python3

        #     from pymycobot import MyArm
        #     from pymycobot import PI_PORT, PI_BAUD
        #     import time

        #     start = time.time()
        #     mycobot = MyArm('%s', %s)

        #     def aging_test():
        #         # fast
                


        #     if __name__ == '__main__':
        #         while True and time.time() - start < 60*30:
        #             aging_test()
        #         self.mycobot.send_angles([0, 0, 0, 0, 0, 0, 0], 35)
        #     """
        #     % (self.prot, 115200)
        # )

        # aging_test_content_sh = textwrap.dedent(
        #     """\
        #     #!/bin/bash
        #     /usr/bin/python3 /home/er/Desktop/aging_test.py
        #     """
        # )

        # aging_test_content_service = textwrap.dedent(
        #     """\
        #     [Unit]
        #     Description=aging-test

        #     [Service]
        #     Type=forking
        #     User=er
        #     Restart=on-failure
        #     RestartSec=2
        #     ExecStart=/home/er/aging_test.sh

        #     [Install]
        #     WantedBy=multi-user.target
        #     """
        # )

        # os.system(
        #     'echo "' + aging_test_content_py + '" >> /home/er/Desktop/aging_test.py'
        # )
        # os.system('echo "' + aging_test_content_sh + '" >> /home/er/aging_test.sh')
        # os.system("sudo chmod +x /home/er/aging_test.sh")
        # os.system(
        #     'echo "'
        #     + aging_test_content_service
        #     + '" >> /home/er/Desktop/aging_test.service'
        # )
        # os.system(
        #     "sudo mv /home/er/Desktop/aging_test.service /etc/systemd/system/aging_test.service"
        # )
        # os.system("sudo systemctl enable aging_test.service")
        # os.system("sudo systemctl start aging_test.service")

    def _calibration_test(self):
        self.write_log_to_Text("开始测试校准.")
        angles = [0, 0, 0, 0]
        test_angle = [-20, 20, 0]
        for i in range(6):
            for j in range(3):
                angles[i] = test_angle[j]
                if i == 1 and angles[i] < 0:
                    angles[i] = 40
                self.mycobot.send_angles(angles, 50)
                time.sleep(0.7)
        self.write_log_to_Text("测试校准结束.")

    def get_serial_port_list(self):
        plist = [
            str(x).split(" - ")[0].strip() for x in serial.tools.list_ports.comports()
        ]
        print(plist)
        self.port_list["value"] = plist
        return plist

    def get_current_time(self):
        """Get current time with format."""
        current_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
        return current_time

    def write_log_to_Text(self, logmsg: str):
        global LOG_NUM
        current_time = self.get_current_time()
        logmsg_in = str(current_time) + " " + str(logmsg) + "\n"  # 换行

        if LOG_NUM <= 18:
            self.log_data_Text.insert(tkinter.END, logmsg_in)
            LOG_NUM += len(logmsg_in.split("\n"))
            # print(LOG_NUM)
        else:
            self.log_data_Text.insert(tkinter.END, logmsg_in)
            self.log_data_Text.yview("end")


if __name__ == "__main__":
    MycobotTest().run()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import json
import struct
import socket
import threading
from pymycobot.mercury_api import MercuryCommandGenerator
from pymycobot.error import calibration_parameters

"""
for mercury x1 robot arms (six-axis or seven-axis), using socket to communicate with the robot
"""


class MercuryArmsSocket(MercuryCommandGenerator):
    def __init__(self, arm, ip, netport=9000, debug=False):
        """
        Arm socket connection
        Args:
            arm: 'left_arm' or 'right_arm'
            ip: ip address
            netport: port
            debug: debug mode
        """
        super(MercuryArmsSocket, self).__init__(debug)
        self.arm = arm
        self.calibration_parameters = calibration_parameters
        self.SERVER_IP = ip
        self.SERVER_PORT = netport
        self.sock = self.connect_socket()
        self.lock = threading.Lock()
        self.lock_out = threading.Lock()
        self.read_threading = threading.Thread(target=self.read_thread, args=("socket", ))
        self.read_threading.daemon = True
        self.read_threading.start()
        self.get_limit_switch()

    def connect_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.SERVER_IP, self.SERVER_PORT))
        return sock

    def open(self):
        self.sock = self.connect_socket()

    def close(self):
        self.sock.close()
        self.sock = None

    def __format(self, command: str) -> bytes:
        send_data = {"command": command, "arm": self.arm}
        data_byter = json.dumps(send_data).encode('utf-8')
        date_length = struct.pack('!I', len(data_byter))
        return b''.join([date_length, data_byter])

    def _write(self, command, method=None):
        log_command = " ".join(map(lambda n: hex(n)[2:], command))
        self.log.debug("_write: {}".format(log_command))
        self.sock.sendall(self.__format(command))



























